#!/usr/bin/perl
#@(#)purecovweb.pl Version: 1.16  Date: 9/10/2009 $
#@(#)(C) Copyright IBM Corporation. 2007, 2009. All Rights Reserved.
#------------------------------------------------------------------------
# Refer to the purecovweb.html file for usage information.
# Also, you can run this script to view the usage info file in text mode:
#    perl purecovweb.pl -help
#------------------------------------------------------------------------
my $cmdFile         = $0;
my ($cmdName)       = $cmdFile =~ m/([^\/\\]+)$/;
my ($dirName)       = $cmdFile =~ m/(.*)[\/\\]/ ? $1 : "";
my $imgDir          = $dirName eq "" ? "$cmdName" : "$dirName/$cmdName";
   $imgDir          =~ s/.pl$/_files/;
my $rev             = "1.16";
my $productName     = "Rational PureCoverage";
my $project         = "";
my $itemWidth       = 60;
my $widthMin        = 20;
my $widthMax        = 128;
my $pureCovTxt      = "";
my $force           = 0;
my $verbose         = 0;
my $doAnnotate      = 0;
my $doInclude       = 0;
my $doExclude       = 0;
my $doSubstitute    = 0;
my $doSearch        = 0;
my $cmdOutOnly      = 0;
my $cmdOut          = "";
my $cmdOutN         = 0;
my %includeList     = ();
my $excludeMatch    = "";
my %excludeList     = ();
my $substituteList  = "";
my %searchList      = ();
my $var             = ();
my $msgCat          = ();
my $truncate        = 'center';
my $destDir         = "";
my $lang            = 'en';  # English
my $thisMsg         = "";
my ($onUnix,$onWin) = &determineOs;
my $host            = &getHostName;
my $user            = &getUserName;
my @allArgs         = defined $ARGV[0] ? @ARGV : ('','');
   $imgDir          =~ s/\//\\/g if $onWin;
my $cmdHelp         = $cmdName;
   $cmdHelp         =~ s/\.pl$/.en.html/;
   $cmdHelp         = $onWin ?  "$imgDir\\$cmdHelp" : "$imgDir/$cmdHelp";
my $langHelp        = "";
my $helpName        = $cmdName;
   $helpName        =~ s/\.pl$/.en.html/;
my $msgFile         = $cmdHelp;
   $msgFile         =~ s/.en.html/.msg.en.txt/;
my $cmdOutHtml      = "$cmdName.out.html";
   $cmdOutHtml      =~ s/\.pl//;
my $iOpt            = $onWin ? 'i' : ''; # Use option i to ignore case on Windows.
my $includeMatch    = "m#\\t(?:Run|Module)\\t#$iOpt";

# Check command line options for language.  Need this to load the correct message file.
my $i = 0;
while (defined $ARGV[$i]) {
    if ($ARGV[$i] =~ m/^-l/) {
        ($lang) = $ARGV[++$i] =~ m/^(\w\w)/;
        $msgFile  =~ s/en\.txt$/$lang.txt/;
        $langHelp = $cmdHelp;
        $langHelp =~ s/en.html$/$lang.html/;
        $helpName =~ s/en.html$/$lang.html/;
        if ($onUnix and $lang ne 'en') {
            $cmdOutOnly = 1;
        }
        last;
    }
    $i++;
}

# Load the message file.  Need this before we do any command line processing.
&loadMsgVar();
#YYmy $srcWarn     = $msgCat{msg36};
my $AdjLines    = $var{var01};  # Adjusted Lines
my $AnnSource   = $var{var02};  # Annotated Source
my $SourceView  = $var{var03};  # Annotated Source View
my $Calls       = $var{var04};  # Calls
my $CantAnn     = $var{var05};  # Cannot Annotate
my $Colors      = $var{var06};  # Colors
my $Coverage    = $var{var07};  # Coverage
my $CovItem     = $var{var08};  # Coverage Item:
my $DeadCode    = $var{var09};  # dead code
my $DeadLines   = $var{var10};  # Dead Lines
my $Directory   = $var{var11};  # Directory
my $DirView     = $var{var12};  # Directory View
my $FileView    = $var{var13};  # File View
my $Folder      = $var{var14};  # Folder
my $FolView     = $var{var15};  # Folder View
my $Function    = $var{var16};  # Function
my $Functions   = $var{var17};  # Functions
my $Hit         = $var{var18};  # hit
my $Hits        = $var{var19};  # Hits
my $Inspected   = $var{var20};  # inspected
my $Line        = $var{var21};  # Line
my $Lines       = $var{var22};  # Lines
my $LinesHit    = $var{var23};  # Lines Hit
my $LinesMissed = $var{var24};  # Lines Missed
my $LinesUnused = $var{var25};  # Lines Unused
my $LinesUsed   = $var{var26};  # Lines Used
my $MachDetail  = $var{var27};  # Machine Details
my $Method      = $var{var28};  # Method
my $Methods     = $var{var29};  # Methods
my $Missed      = $var{var30};  # Missed
my $MissSrc     = $var{var31};  # Missing Source Files
my $ModuleView  = $var{var32};  # Module View
my $NotDefined  = $var{var33};  # Not-Defined
my $Number      = $var{var34};  # Number
my $PartLines   = $var{var35};  # Partially Hit Multi-block Lines
my $ProgInfo    = $var{var36};  # Program Information
my $Run         = $var{var37};  # Run
my $Source      = $var{var38};  # Source
my $Tested      = $var{var39};  # tested
my $Unused      = $var{var40};  # unused
my $Used        = $var{var41};  # used
my $version     = $var{var42};  # version
if ($cmdOutOnly) {
    # These do not translate in a UNIX terminal, so force English.
    $msgCat{msg21} = "%s1 already exists.";
    $msgCat{msg22} = "Do you want to remake it? (yes|no) [no]";
}

# When adding new options, remember to update purecovweb.html (helpName) too.
while (defined $ARGV[0]) {
    my $arg = shift @ARGV;
    if ($arg =~ m/^-a/) {
        $doAnnotate = 1;
        # msg01: <cmd>: Do annotated source: yes
        &ifVerbose(__LINE__,'msg01');

    } elsif ($arg =~ m/^-d/) {
        $destDir = shift @ARGV;
        if (! -d $destDir) {
            print `mkdir $destDir 2>&1`;
            if (! -d $destDir) {
                # msg28: <cmd>: <dir> not found, cannot create.
                &usage(__LINE__,'msg28',$destDir);
            }
        } elsif (! -w $destDir) {
            # msg29: <cmd>: <dir> not writable.
            &cmdMsg(__LINE__,'msg29',$destDir);
        }
        $destDir .= $onUnix ? "/" : "\\";
        # msg02: <cmd>: Destination directory: <dir>
        &ifVerbose(__LINE__,'msg02',$destDir);

    } elsif ($arg =~ m/^-e/) {
        # Options used in the sub filterData
        $doExclude   = 1;
        $doAnnotate  = 1;
        $arg = shift @ARGV;
        for (&splitPaths($arg)) {
            # msg03: <cmd>: Will exclude: <path>
            &ifVerbose(__LINE__,'msg03',$_);
            s/\\/\\\\/g;
            $excludeList{$_} = 1;
            s/\#/\\#/g;
            $excludeMatch .= " or m#\\t$_#$iOpt";
        }

    } elsif ($arg =~ m/^-f/) {
        # msg04: <cmd>: Force: Will replace existing index or default HTML file.
        &ifVerbose(__LINE__,'msg04');
        $force = 1;

    } elsif ($arg =~ m/^-h/) {
        &usage(__LINE__,'msg00');

    } elsif ($arg =~ m/^-i/) {
        # Options used in the sub filterData
        $doInclude   = 1;
        $doAnnotate  = 1;
        $arg = shift @ARGV;
        for (&getPaths($arg)) {
            &ifVerbose(__LINE__,'msg57',$_);
            s/\\/\\\\/g;
            $includeList{$_} = 1;
            s/\#/\\#/g;
            $includeMatch .= " or m#\\t$_#$iOpt";
        }

    } elsif ($arg =~ m/^-l/) {
        # Already processed this option.  Just shift out the arg.
        $arg = shift @ARGV;

    } elsif ($arg =~ m/^-m/) {
        # msg12: <cmd>: Will run in verbose message mode.
        &ifVerbose(__LINE__,'msg12');
        $verbose = 1;

    } elsif ($arg =~ m/^-p/) {
        $project = shift @ARGV;
        # msg60: <cmd>: Using '$project' as the project name.
        &ifVerbose(__LINE__,'msg60',$project);

    } elsif ($arg =~ m/^-so/) {
        # Options used in the sub filterData
        $doSearch    = 1;
        $doAnnotate  = 1;
        $arg = shift @ARGV;
        $arg =~ s/\#/\\#/g;
        $arg =~ s/\\/\\\\/g;
        for (&getPaths($arg)) {
            # msg05: <cmd>: Will search source:
            &ifVerbose(__LINE__,'msg05',$_);
            $searchList{$_} = 1;
        }

    } elsif ($arg =~ m/^-su/) {
        # Options used in the sub filterData
        $doSubstitute = 1;
        $doAnnotate   = 1;
        $arg = shift @ARGV;
        for (&splitPaths($arg)) {
            my ($origPath,$newPath) = split /\|/;
            # msg06: <cmd>: Will substitute "$origPath" with "$newPath."
            &ifVerbose(__LINE__,'msg06',$origPath,$newPath);
            $origPath =~ s/\#/\\#/g;
            $origPath =~ s/\\/\\\\/g;
            $newPath  =~ s/\#/\\#/g;
            $newPath  =~ s/\\/\\\\/g;
            $substituteList .= "s#$origPath#$newPath#g$iOpt; ";
        }

    } elsif ($arg =~ m/^-t/) {
        my $arg2 = shift @ARGV;
        if ($arg2 =~ m/^l/i) {
            $truncate = 'left';
        } elsif ($arg2 =~ m/^r/i) {
            $truncate = 'right';
        } elsif ($arg2 =~ m/^m/i or $arg2 =~ m/^c/i) {
            $truncate = 'center';
        } else {
            # msg09: <cmd>: <arg2> is not a recognized keyword for the <arg> option.
            &cmdMsg(__LINE__,'msg09',$arg2,$arg);
            # msg10: <cmd>: You must use one of these keywords:  right  left  center
            &usage(__LINE__,'msg10',' right  left  center');
        }
        # msg11: <cmd>: Will truncate <truncate>
        &ifVerbose(__LINE__,'msg11',$truncate);

    } elsif ($arg =~ m/^-v/) {
        # Display version and exit.
        &cmdMsg(__LINE__,'msg00',"$version $rev");
        &doExit(0);

    } elsif ($arg =~ m/^-w/) {
        $itemWidth = shift @ARGV;
        if ($itemWidth =~ m/^\d+/) {
            if ($itemWidth < $widthMin) {
                # msg13: <cmd>: Setting <item> width to <min>, the minimum width
                # (<itemWidth> less than $widthMin).
                &ifVerbose(__LINE__,'msg13',$CovItem,$widthMin,$itemWidth,$widthMin);
                $itemWidth = $widthMin;
            } elsif ($itemWidth > $widthMax) {
                # msg14: <cmd>: Setting <item> width to <max>, the maximum width
                # (<itemWidth> greater than <widthMax>).
                &ifVerbose(__LINE__,'msg14',$CovItem,$widthMax,$itemWidth,$widthMax);
                $itemWidth = $widthMax;
            } else {
                # msg15: <cmd>: Width of <item> <width>
                &ifVerbose(__LINE__,'msg15',$CovItem,$itemWidth);
            }
        } else {
            # msg16: <cmd>: The "<arg>" option requires a width number.  "<itemWidth>" is not a number."
            &usage(__LINE__,'msg16',$arg,$itemWidth);
        }
    } elsif ($arg =~ m/^-/) {
        # msg17: <cmd>: Option "<arg>" not recognized.
        &usage(__LINE__,'msg17',$arg);
    } else {
        $pureCovTxt = $arg;
        # msg07: <cmd>: The PureCoverage text file is:
        &ifVerbose(__LINE__,'msg07',$pureCovTxt);
        last;
    }
}

if ($pureCovTxt eq '') {
    # msg18: <cmd>: The $productName text file was not specified.
    &usage(__LINE__,'msg18',$productName);
}
if ($doInclude and $doExclude) {
    # msg19: <cmd>: The "-i" and "-e" options are mutually exclusive.
    &usage(__LINE__,'msg19');
}
my ($pcFunName,$pcFilName,$pcFunPath,$pcFilPath);

if ($doSubstitute) {
    $substituteList =~ s/; $//;
    $doSubstitute = eval "sub { $substituteList }";
} else {
    $doSubstitute = eval "sub { return 1 }";
}

if ($doInclude) {
    $includeMatch =~ s/^ or //;
    $includeMatch = eval "sub { $includeMatch }";
} else {
    $includeMatch = eval "sub { return 1 }";
}

if ($doExclude) {
    $excludeMatch =~ s/^ or //;
    $excludeMatch = eval "sub { $excludeMatch }";
} else {
    $excludeMatch = eval "sub { return 0 }";
}

# Determine if the text file is a PureCoverage text file.  Also, determine
# if the PureCoverage text file is from Windows or UNIX.  Read no more than
# the first hundred lines.  If there is a line with
# "comment\tPureCoverage text data file:" within the first few lines,
# it is Windows.  If there is a line that starts with "^to\t" within the
# first hundred lines, it is UNIX.
unless (open(TXT,"<$pureCovTxt")) {
    # msg48: <cmd>: Failed to open <file> for reading: $!
    &myDie(__LINE__,'msg48',$pureCovTxt,$!);
}
$i = 1;
my $isUnixTxt = -1;  # Assume neither UNIX or Windows to start.
my $isWinTxt  = -1;
while (<TXT>) {
    if (m/^comment\tProgram Name\t/) {
        $isUnixTxt = 0;
        last;
    } elsif (m/^to\t/) {
        $isUnixTxt = 1;
        last;
    }
}

if ($isUnixTxt < 0) {
    # msg20: <cmd>: The "<txt-file>" file does not have any <productName> tags.\n";
    &cmdMsg(__LINE__,'msg20',$pureCovTxt,$productName);
    close TXT;
    &doExit(1);
}
seek TXT,0,0;  # Reset file pointer back to the start.

my ($copyCmd,$missHtml,$sep);
my $htmEnd = "";
if ($onUnix) {
    $htmEnd    = '.html';
    $sep       = '/';
    $imgDir   .= '/';
    $copyCmd   = 'cp';
    $pcFunName = "index.html";
    $missHtml  = "missing.html";

} else {
    $htmEnd    = '.htm';
    $sep       = '\\';
    $imgDir   .= '\\';
    $copyCmd   = 'copy';
    $pcFunName = "default.htm";
    $missHtml  = "missing.htm";
    $Directory = $Folder;
    $imgDir    =~ s/\//\\/g;
}

if ($isUnixTxt) {
    $Hit         = $Used;
    $Missed      = $Unused;
    $LinesHit    = $LinesUsed;
    $LinesMissed = $LinesUnused;
    $pcFilName   = "";
    $isWinTxt    = 0;
} else {
    $pcFilName   = "default2.htm";
    $isWinTxt    = 1;
}

if ($destDir eq '') {
    if ($onUnix) {
        if ($pureCovTxt =~ m/(.*\/)[^\/]+$/) {
            $destDir = $1;
        }
    } else {
        if ($pureCovTxt =~ m/(.*\\)[^\\]+$/) {
            $destDir = $1;
        }
    }
}
$pcFunPath = "$destDir$pcFunName";
$pcFilPath = "$destDir$pcFilName";

if (-e $pcFunPath and not $force) {
    # msg21: <cmd>: <file> already exists
    &cmdMsg(__LINE__,'msg21',$pcFunPath);
    # msg22: Do you want to remake it? (yes|no) [no]
    $thisMsg = &getMsg('msg22');
    print "$thisMsg ";
    my $remake = <STDIN>;
    if ($remake =~ m/^y.*/i) {
        unlink $pcFunPath;
        if (-e $pcFunPath) {
            # msg49: <cmd>: Cannot remove <file>.
            &myDie(__LINE__,'msg49',$pcFunPath);
        }
    } else {
        # msg23: <cmd>: Okay, will not remake this file.  Quitting.
        &myDie(__LINE__,'msg23');
    }
}

my %modList     = ();
my %dirList     = ();
my %filList     = ();
my %srcList     = ();
my %funList     = ();
my %dir2List    = ();
my %htmlDirs    = ();
my %src2List    = ();
my %src3List    = ();
my %srcMissing  = ();
my %fun2List    = ();
my %lineList    = ();
my %partialHit  = ();
my @srcMap      = ();
my @lineHits    = ();
my %labels      = ();
my $sourceFile  = '';
my $pcColor     = "#33CCCC";
my $startJava   = '<script type="text/javascript">';
my $endJava     = '</script>';
my $startCss    = '<style type="text/css">';
my $endCss      = '</style>';
my $font1       = "<font face=\"sans-serif\">";
my $font2       = "<font face=\"sans-serif\" size=+2>";
my $fontRed     = "<font color=red>";
my $fontBlue    = "<font color=blue>";
my $fontMagenta = "<font color=magenta>";
my $startHtml1   = "<html>\n<title>";
my $startHtml2   = "</title>\n<body bgcolor=\"#ffffff\">\n";
my $endHtml     = "</form>\n</body>\n</html>\n\n";
my $br          = "\n";
my $priorHit    = -1;
my $totalInfo   = "";
my $progName    = "";
my $progArgs    = "";
my $progTime    = "";
my $progLang    = "C";
my @cEnvVars    = ();
my @javaEnvVars = ('CLASSPATH');
my $td          = "";
my $imgSrc      = "";
my $pcImg       = "";
my @imgList1    = ();
my @imgList2    = ();
my $imgN        = -1;
my $imgN1       = -1;
my $imgN2       = -1;
my $imgHeight   = 16;
my $imgWidth    = 32;
my $detailLen   = 26;
my $detailSp    = " " x $detailLen;
my $detailInd   = " " x 8;
my $htmlSpace   = "&nbsp; " x 4;
my $startTime   = localtime();
                  # msg50: View $cmdName Usage Information
   $thisMsg     = &getMsg('msg50',$cmdName);
my $helpLink    = "<input type=button value=\"$thisMsg\"";
   $helpLink   .= " onClick=\"openpopup('purecovweb_help','$helpName')\">";
                  # msg51: Show Image Legend
   $thisMsg     = &getMsg('msg51');
my $imgsLink    = "<input type=button value=\"$thisMsg\"";
   $imgsLink   .= " onClick=\"openpopup('purecovweb_help','$helpName#images')\">";

                  # msg52: View List Of Source Files Not Found
my $notFound    = "";
if ($doAnnotate) {
   $thisMsg     = &getMsg('msg52');
   $notFound   .= "<input type=button value=\"$thisMsg\"";
   $notFound   .= " onClick=\"openpopup('purecovweb_help','$missHtml')\">";
}
                  # msg59: View Program Output
my $cmdOutLink  = "";
   $thisMsg     = &getMsg('msg59');
   $cmdOutLink  .= "<input type=button value=\"$thisMsg\"";
   $cmdOutLink  .= " onClick=\"openpopup('purecovweb_help','$cmdOutHtml')\">";
                  # msg53: This file was generated by user "<user>" on machine "<host>"
                  # at <time>, using <cmd>, version <rev>
   $thisMsg     = &getMsg('msg53',$user,$host,$startTime,$cmdName,$rev);
my $details     = "$detailInd $thisMsg";
   $details    .= "$detailInd    $cmdName @allArgs\n\n";
   $details    .= "$detailInd $helpLink $htmlSpace $imgsLink $htmlSpace $notFound $htmlSpace $cmdOutLink\n\n";
                  # msg54: If the columns are not lined up, check your";
                  # browser monospace font is set to a fixed width font";
                  # (e.g. Courier New, Fixedsys, Typewriter).\n\n";
   $thisMsg     = &getMsg('msg54');
   $details    .= "$detailInd $thisMsg\n\n";
#YYmy $epochStart  = 0;
my $indent1     = 3;
my $indent2     = $indent1 * 2;
my $sp0         = "";
my $sp1         = " " x $indent1;
my $sp2         = " " x $indent2;
my $showNoSrc   = 0;
my $annotateNum = 0;
my $annotatePre = "src";
my $noSource    = "";
my @annotateLst = ();
my $dirNum      = 0;
my $nImages     = 0;
my $lastExpCol  = -1; 
my $imageList   = "";
my $imageListC  = "//";
my %images      = ();
my @images      = (
   'expand.jpg',        'collapse.jpg',
   'expand-run.jpg',    'collapse-run.jpg',
   'expand-mod.jpg',    'collapse-mod.jpg',
   'expand-dir.jpg',    'collapse-dir.jpg',
   'expand-src.jpg',    'collapse-src.jpg',
   'expand-no-src.jpg', 'collapse-no-src.jpg',
   'popup-dir.jpg' ,    'popup-fun.jpg',
   'fun.jpg'       ,    'pc.gif',
   'blank.jpg'     ,
);
for (@images) {
    $lastExpCol++ if (m/^expand/ or m/^collapse/);
    $imageListC .= sprintf "  %2d = %-16s", $nImages, $_;
    $images{$_}  = $nImages++;
    $imageList  .= sprintf "  %-21s", "'$_',";
    if ($nImages%6 == 0) {
        $imageList  .= "\n";
        $imageListC .= "\n//";
    }
}

# Settings, depending on the local lanuage.
my ($charSet,$bOn,$bOff,$fontBold);
if ($lang eq 'ja') {
    # Japanese -- Turn off bold; does not work in this charset.
    &setCharSetBold('shift-jis','off');
} else {
    &setCharSetBold('utf-8','on');
}

if ($isUnixTxt) {
    &parseUnixTxtFile;
} else {
    &parseWinTxtFile;
}
close TXT;
if ($totalInfo eq '') {
    # msg24: <cmd>: <file> file does not appear to be a $productName text file.
    &myDie(__LINE__,'msg24',$pureCovTxt,$productName);
}

if ($doAnnotate) {
    # msg55: The following source file could not be found:
    $noSource = &getMsg('msg55');
    my @envVars = ();
    if ($progLang eq 'Java') {
        @envVars = @javaEnvVars;
    } else {
        @envVars = @cEnvVars;
    }
    my ($env,$path);
    for $env (@envVars) {
        if (defined $ENV{$env}) {
            for $path (&getPaths($ENV{$env})) {
                $doSearch = 1;
                $searchList{$path} = 1;
            }
       }
   }
} else {
    # msg56: These pages were generated without producing annotated source:
    $noSource = &getMsg('msg56');
}

my $htmlPrologue1 = "<html>
<head>
<title>$productName: $project</title>
<meta http-equiv=\"Content-Type\" content=\"text/html;charset=$charSet\">
<!-- This file was generate using the $cmdName command, version $rev -->
<!--   $cmdName @allArgs -->
$startCss
.hide {display:none}
.show {display:block}
$endCss
$startJava
function openpopup(name,url) {
  smallwin = window.open(url,name,\"location=yes,menubar=yes,resizable=yes,scrollbars=yes,status=yes,titlebar=yes,toolbar=yes,directories=no,height=800,width=1200\");
  smallwin.focus();
}
function showStats(name,calls,funused,fused,lunused,lused,lpct,src) {
  hitOrMiss = (funused == '$Missed') ? '$Missed' : '$Hit';
  mesg  = '$noSource\\n\\t'+src+'\\n';
  mesg += '$Function: '+name+'\\n';
  mesg += '$Calls:\\t'+calls+'\\n';
  mesg += '$Function:\\t'+hitOrMiss+'\\n';
  mesg += '$LinesMissed:\\t'+lunused+'\\n';
  mesg += '$LinesHit:\\t'+lused+'\\n';
  mesg += '% $LinesHit:\\t'+lpct;
  window.alert(mesg);
}
function hideShow(id) {
  section = 'sec'+id;
  thisSec = document.getElementById(section);
  if (thisSec.className == 'show') {
    thisSec.className = 'hide';
  } else {
    thisSec.className = 'show';
  }
  if (imgMap[id] == 0) {
    document.images[id].src = imageB[id].src;
    imgMap[id] = 1;
  } else {
    document.images[id].src = imageA[id].src;
    imgMap[id] = 0;
  }
}
$endJava
</head>
<body bgcolor=white>
<font face=\"monospace\">
$startCss
a:link, a:visited, a:active {color:black; text-decoration:none}
a:hover {background-color:$pcColor; color:black$fontBold}
img {border:0;valign:center}
ul {text-align:left;  width:auto; margin:0; padding:0; list-style:none}
td {text-align:center; font:monospace}
th {text-align:center; font:sans-serif}
$endCss\n";

$imageListC =~ s/.jpg//g;
my $loadImages = "$startJava
$imageListC
var imageList = [
$imageList
];
var imgMap = [
_IMAGE_MAP_
];
var nImage = imgMap.length;
var imageA = new Array(nImage);
var imageB = new Array(nImage);
// Load all images.  Set alternate imageB for images 0..$lastExpCol.
for (i=0; i<nImage; i++) {
  var imgI = imgMap[i];
  imageA[i] = new Image($imgWidth,$imgHeight);
  imageA[i].src = imageList[imgI];
  imageB[i] = new Image($imgWidth,$imgHeight);
  imgMap[i] = 0;
  if (imgI <= $lastExpCol) {
    // Image 0 to $lastExpCol have an alternate image.
    imageB[i].src = imageList[imgI^1];
  } else {
    // Image 12 and above do not have an alternate.
    imageB[i].src = imageList[imgI];
  }
}
$endJava";

my $htmlEpilogue  = "</pre></ul>\n$loadImages\n</body>\n</html>\n";
my $maxLen0 = $itemWidth;
my $maxLen1 = $maxLen0 - $indent1;
my $maxLen2 = $maxLen0 - $indent2;
my $spaces  = " " x $maxLen0;

# Open the function view HTML file.  This is the default or index file.
unless (open(FUN,">$pcFunPath")) {
    # msg31: <cmd>: Cannot create: <file>: <status>
    &myDie(__LINE__,'msg31',$pcFunPath,$!);
}
if ($onWin) {
    # On Windows, open the 'file view' HTML too.
    unless (open(FIL,">$pcFilPath")) {
        # msg31: <cmd>: Cannot create: <file>: <status>
        &myDie(__LINE__,'msg31',$pcFilPath,$!);
    }
}

# Remove any prior directory listings.
my $i;
for ($i=1; -e "${destDir}dir$i$htmEnd"; $i++) {
    unlink "${destDir}dir$i$htmEnd";
}

if ($isUnixTxt) {
    &createHtml(1,\*FUN);
    close FUN;
    &processUnixSource if ($doAnnotate);
} else {
    &createHtml(1,\*FUN);
    close FUN;
    &createHtml(0,\*FIL);
    close FIL;
    &processWinSource if ($doAnnotate);
}
&processMissing;

# Copy images and help file.
for (@images,$helpName) {
    if (! -e "$destDir$_") {
        &cmdMsg(__LINE__,'msg00',"$copyCmd $imgDir$_ $destDir$_");
        `$copyCmd $imgDir$_ $destDir$_\n`;
    }
}
# msg25: <cmd>: Created: <file>
&ifVerbose(__LINE__,'msg25',$pcFunPath);
if ($onWin) {
    # msg25: <cmd>: Created: <file>
    &ifVerbose(__LINE__,'msg25',$pcFilPath);
}
my $endTime = localtime();
# msg26: <cmd>: Started: <start-time>
&ifVerbose(__LINE__,'msg26',$startTime);
# msg27: <cmd>: Ended:   <start-time>
&ifVerbose(__LINE__,'msg27',$endTime);
&ifVerbose(__LINE__,'msg00',"$version $rev");
&doExit(0);

# End of main program.

#------------------------------------------------------------------------
sub doExit
{
    my $outFile = "$destDir$cmdOutHtml";
    if (open(CMDHTM,">$outFile")) {
        print CMDHTM $startHtml1;
        print CMDHTM "$cmdName - $ProgInfo";
        print CMDHTM $startHtml2;
        print CMDHTM "<meta http-equiv='Content-Type' content='text/html;charset=$charSet'>\n"
            if ($lang ne 'en');
        print CMDHTM "$user\@$host\: $cmdFile @allArgs</font>\n<hr><pre>\n";
        if ($cmdOutN) {
            print CMDHTM $cmdOut;
        }
        print CMDHTM $endHtml;
        close CMDHTM;
        close CMDHTM;
        my $outMsg = $lang eq 'en' ? " Output written to:" : "";
        printf "%s:%04d-00:%s %s\n",$cmdName,__LINE__,$outMsg,$outFile;
    } else {
        print "XX failed to open: $outFile, $!\n";
    }
    exit $_[0];
}

#------------------------------------------------------------------------
sub parseWinTxtFile
{
    my ($line,$hits,$pHits,$entry,$len,$label);
    my $sourceFile  = "";
    my $noFilters = 1;
    my $line;
    while (<TXT>) {
        chomp;
        if (m/^CoverageData\t(.*)/) {
            unless (&$includeMatch) {
                next;
            }
            if (&$excludeMatch) {
                next;
            }
            &$doSubstitute;
            ($covData,$name,$type,$parentSrc,$parentMod,$funUnused,$funUsed,$funTotal,$funPct,$linUnused,$linUsed,$linTotal,$linPct,$srcLine,$calls) = split /\t/,$_;
            $entry = "$calls|$funUnused|$funUsed|$funPct|$linUnused|$linUsed|$linPct";
            if ($type eq 'Function') {
                ($parentDir) = $parentSrc =~ m|(.*)[\\/]|;
                $funList{$parentMod}{$parentDir}{$parentSrc}{$name} = $entry;
                $fun2List{$parentDir}{$parentSrc}{$name} = $entry;
                if ($doAnnotate) {
                    $src3List{$parentSrc} = $annotatePre . ++$annotateNum;
                    if ($name =~ m/(\S+)\:\:(\S+)\s*\(/) {
                        my $object = &getObject($1,$2);
                        $label     = "$parentSrc|$object";
                        $labels{$label} = 1;
                    }
                }
            } elsif ($type eq 'SourceFile') {
                $srcList{$parentMod}{$parentSrc}{$name} = $entry;
                $src2List{$parentSrc}{$name} = $entry;
            } elsif ($type eq 'Directory') {
                # Uncomment to Skip unknown directories.
                next if ($name eq '(Unknown Directory)' and $funTotal == 0 and $linTotal == 0);
                $dirList{$parentMod}{$name} = $entry;
                $dir2List{$name} = &totalDir($name,$entry);
            } elsif ($type eq 'Module') {
                $modList{$name} = $entry;
            } elsif ($type eq 'Run') {
                $totalInfo = $entry;
            }
        } elsif (m/^comment\t([^\t]+)\t(.*)/) {
            if (m/(case.*ive)\t(delete|hide)(module|file|function)\t(.*abled)\t(.*)/i) {
                if ($noFilters ) {
                    $details  .= "\n$detailInd <u>${bOn}Filters:$bOff</u>\n";
                    $noFilters = 0;
                }
                my $case = sprintf "%-16s",$1;
                my $enab = sprintf "%-10s",$4;
                $td = $2 . $3 . ':' . $detailSp;
                $td = substr $td,0,$detailLen;
                $details .= "$detailInd $td    $enab  $case  $5\n";
            } else {
                my ($cTag1,$cTag2) = ($1,$2);
                $td = $cTag1 . ':' . $detailSp;
                if ($cTag1 =~ m/Jn/) {
                    &setCharSetBold('shift-jis','off');
                    $cTag1 = 'Start Time';
                } elsif ($cTag1 =~ m/I/) {
                    &setCharSetBold('shift-jis','off');
                    $cTag1 = 'Stop Time';
                }
                $td = substr $td,0,$detailLen;
                $details .= "$detailInd $td    $cTag2\n";
                if ($cTag1 eq 'Program Name') {
                    $progName = $cTag2;
                    $project  = $cTag2;
                } elsif ($cTag1 eq 'Program Arguments') {
                    $progArgs  = $cTag2;
                } elsif ($cTag1 eq 'Language') {
                    $progLang = $cTag2;
                    if ($progLang eq 'Java') {
                        $Function    = $Method;
                        $Functions   = $Methods;
                    }
                } elsif ($cTag1 eq 'Stop Time') {
                    $progTime = $cTag2;
                #} elsif ($cTag1 eq 'Start Time') {
                #    # Example: Start Time: 03/20/07 11:58:06
                #    # Mon=$1, Day=$2, Year=$3, Hour=$4, Minute=$5, Second=$6
                #    my $timeInfo = $cTag2;
                #    if ($timeInfo =~ m/(\d\d)\/(\d\d)\/(\d{2,4})\s+(\d\d):(\d\d):(\d\d)/) {
                #        my $yr;
                #        my $y3 = $3;
                #        if ($y3 =~ m/\d{4}/) {
                #            $yr = $y3;
                #        } elsif ($y3 =~ m/^9\d/) { 
                #            # The year is 199x.  PureCoverage did not exist before 1993.
                #            $yr = "19$y3";
                #        } else {
                #            # Take the first 2 digits of the current 4-digit year and
                #            # append the 2-digit year, $y3.
                #            ($yr) = (localtime())[5] + 1900;
                #            $yr =~ s/..$//;
                #            $yr .= $y3;
                #        }
                #        $epochStart = &getEpoch($6,$5,$4,$2,$1,$yr);
                #    }
                }
            }
        } elsif (m/^SourceLines\t/) {
            &$doSubstitute;
            ($sourceFile) = $1 if (m/^SourceLines\t(.*)\t/);
        } elsif (m/^\t(\d+)\.\d+\t(\d+)$/) {
            $line = $1;
            $hits = $2;
            if (defined $lineList{$sourceFile}{$line}) {
                $pHits = $lineList{$sourceFile}{$line};
                $lineList{$sourceFile}{$line} = $hits if ($hits > $pHits);
                if (($pHits == 0 and $hits > 0) or ($pHits > 0 and $hits == 0)) {
                    $partialHit{$sourceFile}{$line} = 1;
                }
            } else {
                $lineList{$sourceFile}{$line} = $hits;
            }
        }
    }
}

#------------------------------------------------------------------------
sub parseUnixTxtFile
{
    # Summary of keys and field meanings:  (each field is tab-separated in the file)
    #
    # to unused-blocks blocks unused-lines lines unused-funcs funcs files dirs
    # di name unused-blocks blocks unused-lines lines unused-funcs funcs files
    # fi name unused-blocks blocks unused-lines lines unused-funcs funcs sessions
    # fu name unused-blocks blocks unused-lines lines calls
    # li line_no hits
    # Currently, this program ignores:
    # bl line_no hits
    # efu name
    # efi name
    # edi name
    #
    # Example:
    #   to  8498    10496   6982    9238    227     359     13      1
    #   di  /nfs/u40/smcvey/apps/qhull/     8498    10496   6982    9238    227 359     13
    #   fi  io.c    2359    2569    1804    1974    63      71      1
    #   fu  qh_printbegin   255     255     181     181     0
    #   li  1222    0

    my ($type,$blkUnused,$blocks,$linUnused,$lines,$funUnused,$funcs,$files,$dirs,$sessions);
    my ($lineNo,$hits,$entry,$len,$blkPct,$sPath);
    my $label  = "";
    my $skipIt = 0;
    while (<TXT>) {
        chomp;
        if (m/^li\t/) {
            next if $skipIt;
            ($type,$lineNo,$hits) = split /\t/,$_;
            $sPath = "$parentDir/$parentSrc";
            $lineList{$sPath}{$lineNo} = $hits;
            if ($label ne '') {
                $labels{"$label|$lineNo"}    = $parentMod;
                $labels{"$label|$parentMod"} = $lineNo;
                $label = "";
            }
        } elsif (m/^di\t/) {
            unless (&$includeMatch) {
                # Skip everything until the next di.
                $skipIt = 1;
                next;
            }
            if (&$excludeMatch) {
                # Skip everything until the next di.
                $skipIt = 1;
                next;
            }
            $skipIt = 0;
            s/\*\*//g;
            &$doSubstitute;
            ($type,$name,$blkUnused,$blocks,$linUnused,$lines,$funUnused,$funcs,$files) =
                split /\t/,$_;
            $name      =~ s/\/$//;
            $parentDir = $name;
            $funUsed   = $funcs - $funUnused;
            $linUsed   = $lines - $linUnused;
            $funPct    = $funcs ? sprintf "%.2f",$funUsed/$funcs*100 : '0.00';
            $linPct    = $lines ? sprintf "%.2f",$linUsed/$lines*100 : '0.00';
            $entry     = "$files|$funUnused|$funUsed|$funPct|$linUnused|$linUsed|$linPct";
            $dir2List{$name} = $entry; 
            unless ($project) {
                $project = ($name =~ m/no_directory/) ? '&nbsp;' : $name;
            }
        } elsif (m/^fi\t/) {
            next if $skipIt;
            ($type,$name,$blkUnused,$blocks,$linUnused,$lines,$funUnused,$funcs,$sessions) =
                split /\t/,$_;
            $funUsed   = $funcs - $funUnused;
            $linUsed   = $lines - $linUnused;
            $funPct    = $funcs ? sprintf "%.2f",$funUsed/$funcs*100 : '0.00';
            $linPct    = $lines ? sprintf "%.2f",$linUsed/$lines*100 : '0.00';
            $entry     = "$sessions|$funUnused|$funUsed|$funPct|$linUnused|$linUsed|$linPct";
            $parentSrc = $name;
            $src2List{$parentDir}{$name}  = $entry;
            $src3List{"$parentDir/$name"} = $annotatePre . ++$annotateNum;
        } elsif (m/^fu\t/) {
            next if $skipIt;
            ($type,$name,$blkUnused,$blocks,$linUnused,$lines,$calls) = split /\t/,$_;
            ($funUsed,$funUnused) = $calls ? (1,0) : (0,1);
            $blkPct    = " ";
            $linUsed   = $lines  - $linUnused;
            $linPct    = $lines  ? sprintf "%.2f",$linUsed/$lines*100 : '0.00';
            $parentMod = $name;
            $entry     = "$calls|$funUnused|$funUsed|$blkPct|$linUnused|$linUsed|$linPct";
            $fun2List{$parentDir}{$parentSrc}{$name} = $entry;
            if ($doAnnotate) {
                $label = "$parentDir|$parentSrc";
            }
        } elsif (m/^efu\t/) {
            next if $skipIt;
            # To turn off coloring.
            $lineNo++;
            $sPath = "$parentDir/$parentSrc";
            $lineList{$sPath}{$lineNo} = -1;
        } elsif (m/^to\t/) {
            ($type,$blkUnused,$blocks,$linUnused,$lines,$funUnused,$funcs,$files,$dirs) =
                split /\t/,$_;
            $funUsed   = $funcs - $funUnused;
            $linUsed   = $lines - $linUnused;
            $funPct    = $funcs ? sprintf "%.2f",$funUsed/$funcs*100 : '0.00';
            $linPct    = $lines ? sprintf "%.2f",$linUsed/$lines*100 : '0.00';
            $totalInfo = " |$funUnused|$funUsed|$funPct|$linUnused|$linUsed|$linPct";
        #YY} elsif (m/PureCoverage .* created (\d\d)-(\w{3})-(\d{4}) (\d\d):(\d\d):(\d\d) ([AP]M)/i) {
            # Currently this is commented out since the time stamp in this file is the export date,
            # not the date of the PureCoverage run.
            # Example: This is a PureCoverage export file, created 19-Mar-2007 04:22:37 PM
            # 1=day of month, 2=Month, 3=Year, 4=Hour, 5=Minute, 6=Second, 7=AM/PM
            #my ($dy,$mn,$yr,$hr,$mi,$se,$ap) = ($1,$2,$3,$4,$5,$6,$7);
            #my $hr += 12 if ($ap =~ m/pm/i);
            #print "started:  $yr/$mn/$dy at $hr:$mi:$se "; #XX
            #$epochStart = &getEpoch($se,$mi,$hr,$dy,$mn,$yr);
            #print "epoch: $epochStart\n"; #XX
        }
    }
}

#------------------------------------------------------------------------
sub createHtml
{
    my $doFun  = $_[0];
    my $doFil  = $doFun ? 0 : 1;
    my $fd1    = $_[1];
    my @imgSav = ();
    @imgList1  = ();
    $imgN1     = -1;
    &setImgSrc(1,'no','pc');
    $pcImg = $imgSrc;

    my ($run,$win1,$htmlPrologue2);
    if ($isUnixTxt) {
        $win1 = 0;
        &setImgSrc(1,'collapse','expand');
        $htmlPrologue2  = "$sp1$imgSrc <a href=\"javascript:void(0);\" onClick=\"hideShow('$imgN1')\">";
        $htmlPrologue2 .= "$ProgInfo</a>\n<ul class=\"hide\" id=\"sec$imgN1\">\n";
        $htmlPrologue2 .= "$details</ul>";
        $run  = "Total Coverage";
        print $fd1 "$htmlPrologue1
<table width=100% border=1><tr>
<th width=3% bgcolor=$pcColor>$pcImg</th>
<th width=97% bgcolor=$pcColor>
$font1$productName: $project</td>
</tr></table><pre>
$htmlPrologue2";
        @imgSav = @imgList1;
        &unixHeader($fd1,1,$sp1,$maxLen1);
        &addIt($fd1,1,$sp1,$maxLen1,'show','run',$run,$totalInfo);
    } else {
        &setImgSrc(1,'collapse','expand');
        $htmlPrologue2  = "$sp1$imgSrc <a href=\"javascript:void(0);\" onClick=\"hideShow('$imgN1')\">";
        #MC
        $htmlPrologue2 .= "$ProgInfo --- $MachDetail</a>\n<ul class=\"hide\" id=\"sec$imgN1\">\n";
        $htmlPrologue2 .= "$details</ul>";
        $run = "$Run \@ $progTime $progName $progArgs";
        if ($doFun) {
            $win1 = 1;
            print $fd1 "$htmlPrologue1
<table width=100% border=1><tr>
<th width=15%>
$font2$ModuleView</td>
<th width=15% bgcolor=$pcColor>
$font1<a href=$pcFilName><u>$FileView</u></a></td>
<th width=3% bgcolor=$pcColor>$pcImg</th>
<th width=67% bgcolor=$pcColor>
$font1$productName: $project</td>
</tr></table><pre>
$htmlPrologue2";
            @imgSav = @imgList1;
            &winHeader($fd1,1,$sp2,$maxLen2);
        } else {
            $win1 = 0;
            print $fd1 "$htmlPrologue1
<table width=100% border=1><tr>
<th width=15% bgcolor=$pcColor>
$font1<a href=$pcFunName><u>$ModuleView</u></a></td>
<th width=15%>
${font2}$FileView$bOff</td>
<th width=3% bgcolor=$pcColor>$pcImg</th>
<th width=67% bgcolor=$pcColor>
$font1$productName: $project</td>
</tr></table><pre>
$htmlPrologue2";
            @imgSav = @imgList1;
            &winHeader($fd1,1,$sp2,$maxLen2);
        }
        &addIt($fd1,1,$sp0,$maxLen0,'show','run',$run,$totalInfo);
    }

    my ($mod,$dir,$src,$fun,$srcName,$dirName,$pcDirHtm,$dirPro,$dirI);
    my ($doThisFun,$srcLink);
    for $mod (&getKeys('modList',$win1)) {
        &addIt($fd1,1,$sp1,$maxLen1,'hide','mod',$mod,$modList{$mod}) if ($win1);
        for $dir (&getKeys('dirList',$mod)) {
            if ($dir =~ m/(.*)\\/) {
                $dirName = $1;
            } else {
                $dirName = $dir;
            }
            $htmlDirs{$dir} = $dirI = ++$dirNum;
            $pcDirHtm  = "${destDir}dir$dirI$htmEnd";
            if ($win1) {
                &addIt($fd1,1,$sp2,$maxLen2,'hide',"lnk|popup-dir|dir$dirI|dir|",$dir,$dirList{$mod}{$dir});
            } else {
                &addIt($fd1,1,$sp2,$maxLen2,'show',"lnk|popup-dir|dir$dirI|dir|",$dir,$dir2List{$dir});
            }
            next if (-e $pcDirHtm);
            my $fd2 = &createDirHtml($pcDirHtm,$dir,$htmlPrologue2);
            @imgList2 = @imgSav;
            $imgN2    = $#imgSav;
            if ($isUnixTxt) {
                &unixHeader($fd2,2,$sp0,$maxLen0);
            } else {
                &winHeader($fd2,2,$sp0,$maxLen0);
            }
            &addIt($fd2,2,$sp1,$maxLen1,'show','dir',$dir,$dir2List{$dir});
            for $src (&getKeys('srcList',$mod,$dir)) {
                ($srcName)  = $src =~ m|([^\\/]+)$|;
                $sourceFile = $isUnixTxt ? "$dir/$srcName" : $src;
                $sourceFile =~ s/([^\w])/\\$1/g;
                $showNoSrc  = 0;
                $doThisFun  = 0;
                $srcLink    = "";
                if ($doAnnotate) {
                    if ($isUnixTxt) {
                        if (-e "$dir/$srcName") {
                            $doThisFun  = 1;
                            $srcLink = $src3List{"$dir/$srcName"};
                        } else {
                            $showNoSrc  = 1;
                        }
                    } else {
                        if (-e $src) {
                            $doThisFun  = 2;
                            $srcLink = $src3List{$src};
                        } else {
                            $showNoSrc  = 1;
                        }
                    }
                }
                if ($win1) {
                    &addIt($fd2,2,$sp2,$maxLen2,'hide','src',$srcName,$srcList{$mod}{$dir}{$src});
                } else {
                    &addIt($fd2,2,$sp2,$maxLen2,'hide','src',$srcName,$src2List{$dir}{$src});
                }
                for $fun (&getKeys('funList',$mod,$dir,$src)) {
                    my $type = 'fun';
                    if ($doThisFun == 1) {
                        $type = "lnk|fun|$srcLink|$fun|";
                    } elsif ($doThisFun == 2) {
                        if ($fun =~ m/(\S+)::(\S+)\s*\(/) {
                            my $object = &getObject($1,$2);
                            $type = "lnk|fun|$srcLink|$object|";
                        } else {
                            $type = "lnk|fun|$srcLink|$fun|";
                        }
                    }
                    if ($win1) {
                        &addIt($fd2,2,$sp2,$maxLen2,'hide',$type,$fun,$funList{$mod}{$dir}{$src}{$fun});
                    } else {
                        &addIt($fd2,2,$sp2,$maxLen2,'hide',$type,$fun,$fun2List{$dir}{$src}{$fun});
                    }
                }
                print $fd2 "</ul>";
            }
            print $fd2 "</ul>";
            &prntEpilogue(2,$fd2);
            close $fd2;
        }
        print $fd1 "</ul>";
    }
    &prntEpilogue(1,$fd1);
}

#------------------------------------------------------------------------
sub createDirHtml
{
    my ($dirHtml,$dir,$htmlPro2) = @_;
    unless (open(DIR,">$dirHtml")) {
        # msg31: <cmd>: Cannot create: <file>: <status>
        &myDie(__LINE__,'msg31',$dirHtml,$!);
    }
    # msg25: <cmd>: Created: <file>
    &ifVerbose(__LINE__,'msg25',$dirHtml);
    (my $prologue = $htmlPrologue1) =~ s/: $project</: $dir</;
    print DIR $prologue;
    print DIR "<table width=100% border=1><tr>\n";
    print DIR "     <th width=15%>$font2$DirView</th>\n";
    print DIR "     <th width=3% bgcolor=$pcColor>$pcImg</th> ";
    print DIR "     <th width=82% bgcolor=$pcColor>$font1$productName: ";
    print DIR "$project<br>$Directory: $dir</th>\n";
    print DIR "</tr></table><pre>";
    print DIR $htmlPro2;
    return \*DIR;
}

#------------------------------------------------------------------------
sub processWinSource
{
    # Process files for annotated source.
    my ($header1,$header2,$inputSrc,$foundSrc,$winJava);
    $header1  = "<table width=100% border=1><tr>\n";
    $header1 .= "     <th width=15%>$font2$SourceView</th>\n";
    $header1 .= "     <th width=3% bgcolor=$pcColor>$pcImg</th> ";
    $header1 .= "     <th width=82% bgcolor=$pcColor>$font1$productName: ";
    $header1 .= "$project<br>$Source: ";
    $header2  = "</th>\n</tr></table>";
    $header2 .= "<pre> ${bOn}$Colors:";
    $header2 .= " ${fontBlue}$LinesHit</font>,";
    $header2 .= " ${fontRed}$LinesMissed</font>,";
    $header2 .= " $DeadLines,";
    $header2 .= " ${fontMagenta}$PartLines</font>";
    $header2 .= "$bOff\n";
    $winJava  = "<form method=post name=annotated action=\"\">
$startJava
function jumpToObject() {
  var dest = document.annotated.obj.options[document.annotated.obj.selectedIndex].value;
  if (dest != '') {
    dest = '#'+dest;
    document.location.href = dest;
  }
}
$endJava";

    for $inputSrc (sort keys %src3List) {
        if (! -e $inputSrc) {
            $foundSrc = &searchSrc($inputSrc);
            if ($foundSrc ne '') {
                $inputSrc = $foundSrc;
            } else {
                $srcMissing{$inputSrc} = 1;
                next;
            }
        }
        unless (open (INSRC,"<$inputSrc")) {
            # msg30: <cmd>: Cannot read <file>: <status>
            &cmdMsg(__LINE__,'msg30',$inputSrc,$!);
            next;
        }
        my $outputNam = $src3List{$inputSrc} . $htmEnd;
        my $outputSrc = "$destDir$outputNam";
        unless (open (OUTSRC,">$outputSrc")) {
            # msg31: <cmd>: Cannot create: <file>: <status>
            &cmdMsg(__LINE__,'msg31',$outputSrc,$!);
            close INSRC;
            next;
        }
        # msg32: <cmd>: Annotate: <input-file>
        &ifVerbose(__LINE__,'msg32',$inputSrc);
        # msg25: <cmd>: Created: <output-file>
        &ifVerbose(__LINE__,'msg25',$outputSrc);
        my $lnum      = 0;
        my $htmlSrc   = "";
        my $objSel    = "<select name=obj size=1 onChange=\"jumpToObject();\">";
        $objSel      .= "<option value=\"\" selected></option>\n";
        my %objects   = ();
        my $nObj      = 0;
        $priorHit     = -1;
        my $object;
        &setSrcMapLineHits($inputSrc);
        while (<INSRC>) {
            chomp;
            $lnum++;
            if (m/(\S+)::(\S+)(\([^\)]*\))/) {
                $object = &getObject($1,$2);
                my $label = "$inputSrc|$object";
                if (defined $labels{$label}) {
                    $objects{"$1::$2$3"} = $object;
                    $nObj++;
                } else {
                    $object = "";
                }
            } else {
                $object = "";
            }
            $htmlSrc .= &prntWin($inputSrc,$object,$lnum,$_);
        }
        if ($nObj) {
            for (sort keys %objects) {
                $objSel .= "<option value=\"$objects{$_}\">$_</option>\n";
            }
            $object = "$Functions: $objSel</select>"; 
        } else {
            $object = "";
        }
        print OUTSRC $startHtml1;
        print OUTSRC "$AnnSource -- $inputSrc";
        print OUTSRC $startHtml2;
        print OUTSRC $winJava;
        print OUTSRC $header1;
        print OUTSRC $inputSrc;
        print OUTSRC $header2;
        #YYprint OUTSRC " $fontRed$srcWarn.</font>\n" if (&checkModTime($inputSrc));
        print OUTSRC "<hr>";
        print OUTSRC "$bOn   $Line   $bOff|$bOn  $Line  $bOff|$bOn $object$bOff\n";
        print OUTSRC "$bOn $Coverage $bOff|$bOn $Number $bOff|$bOn $AnnSource -- $inputSrc$bOff\n";
        print OUTSRC "<hr>";
        print OUTSRC $htmlSrc;
        print OUTSRC $endHtml;
        close INSRC;
        close OUTSRC;
    }
}

#------------------------------------------------------------------------
sub processUnixSource
{
    # Process files for annotated source.
    my ($header1,$header2,$inputSrc,$outputSrc);
    my ($dir,$src,$lnum);
    $header1  = "<table width=100% border=1><tr>\n";
    $header1 .= "     <th width=15%>$font2$SourceView</th>\n";
    $header1 .= "     <th width=3% bgcolor=$pcColor>$pcImg</th> ";
    $header1 .= "     <th width=82% bgcolor=$pcColor>$font1$productName: ";
    $header1 .= "$project<br>$Source: ";
    $header2  = "</th>\n";
    $header2 .= "</tr></table>";
    $header2 .= "<pre> ${bOn}$Colors:";
    $header2 .= " ${fontBlue}$LinesUsed</font>,";
    $header2 .= " ${fontRed}$LinesUnused</font>,";
    $header2 .= " $DeadLines   D I T: D=$DeadCode, I=$Inspected, T=$Tested";
    $header2 .= "$bOff\n";

    for $dir (sort keys %dir2List) {
        for $src (sort keys %{$src2List{$dir}}) {
            $inputSrc  = "$dir/$src";
            $outputSrc = $destDir . $src3List{"$dir/$src"} . $htmEnd;
            if (! -e $inputSrc) {
                $foundSrc = &searchSrc($inputSrc);
                if ($foundSrc ne '') {
                    $inputSrc = $foundSrc;
                } else {
                    $srcMissing{$inputSrc} = 1;
                    next;
                }
            }
            unless (open (INSRC,"<$inputSrc")) {
                # msg30: <cmd>: Cannot read <file>: <status>
                &cmdMsg(__LINE__,'msg30',$inputSrc,$!);
                next;
            }
            unless (open (OUTSRC,">$outputSrc")) {
                # msg31: <cmd>: Cannot create: <file>: <status>
                &cmdMsg(__LINE__,'msg31',$outputSrc,$!);
                close INSRC;
                next;
            }
            # msg32: <cmd>: Annotate: <input-file>
            &ifVerbose(__LINE__,'msg32',$inputSrc);
            # msg25: <cmd>: Created: <output-file>
            &ifVerbose(__LINE__,'msg25',$outputSrc);
            print OUTSRC $startHtml1;
            print OUTSRC "$AnnSource -- $inputSrc";
            print OUTSRC $startHtml2;
            print OUTSRC "$header1$inputSrc$header2\n";
            #YYprint OUTSRC " $fontRed$srcWarn.</font>\n" if (&checkModTime($inputSrc));
            print OUTSRC "<hr>";
            print OUTSRC "$bOn  $Line $bOff|${bOn}D I T$bOff|$bOn  $Hits $bOff|";
            print OUTSRC "$bOn $AnnSource -- $inputSrc$bOff\n";
            print OUTSRC "<hr>";
            $priorHit = -1;
            &setSrcMapLineHits($inputSrc);
            $lnum     = 0;
            while (<INSRC>) {
                chomp;
                $lnum++;
                &prntUnix($dir,$src,$lnum,$_);
            }
            print OUTSRC $endHtml;
            close INSRC;
            close OUTSRC;
        }
    }
}

#------------------------------------------------------------------------
sub searchSrc
{
    return "" unless $doSearch;
    my $srcFile   = $_[0];
    my ($srcName) = $srcFile =~ m/([^\/\\]+)$/;
    for (keys %searchList) {
        if (-e "$_$sep$srcName") {
            return "$_$sep$//srcName";
        }
    }
    return "";
}

#------------------------------------------------------------------------
sub processMissing
{
    return unless ($doAnnotate);
    unless (open (MISS,">$destDir$missHtml")) {
        # msg31: <cmd>: Cannot create: <file>: <status>
        &cmdMsg(__LINE__,'msg31',"$destDir$missHtml",$!);
        return;
    }
    my ($n,$miss,$incl,$excl,$show,$header1,$header2);
    &setImgSrc(1,'no','pc');
    $pcImg = $imgSrc;
    print MISS $startHtml1;
    print MISS $MissSrc;
    print MISS $startHtml2;
    print MISS "<table width=100% border=1><tr>\n";
    print MISS "     <th width=3% bgcolor=$pcColor>$pcImg</th> ";
    print MISS "     <th width=97% bgcolor=$pcColor>$font1$productName: ";
    print MISS "$project<br>$MissSrc ($CantAnn)";
    print MISS "</th>\n</tr></table><ul>\n";
    $n = 1;
    $miss = scalar keys %srcMissing;
    if ($miss <= 0) {
        # msg58: There are no missing source files -- all source files were found.
        $thisMsg = &getMsg('msg58');
        print MISS "$thisMsg\n";
    } elsif ($doExclude) {
        # msg44: These path names were requested to be excluded.
        $thisMsg = &getMsg('msg44');
        print MISS "$thisMsg\n";
        # msg45: Source files containing these path names were not searched for:
        $thisMsg = &getMsg('msg45');
        print MISS "  $thisMsg<ul>$fontRed\n";
        for $excl (keys %excludeList) {
            $excl =~ s/\\\\/\\/g;
            print MISS "$excl<br>\n";
        }
        print MISS "</font></ul><hr>\n";
        if ($onWin) {
            # Ignore case when matching pathnames on Windows.
            for $miss (sort keys %srcMissing) {
                $show = 1;
                for $excl (keys %excludeList) {
                    if ($miss =~ m/$excl/i) {
                        $show = 0;
                        last;
                    }
                }
                print MISS $n++, ") $miss<br>\n" if $show;
            }
        } else {
            for $miss (sort keys %srcMissing) {
                $show = 1;
                for $excl (keys %excludeList) {
                    if ($miss =~ m/$excl/) {
                        $show = 0;
                        last;
                    }
                }
                print MISS $n++, ") $miss<br>\n" if $show;
            }
        }
    } elsif ($doInclude) {
        # msg46: These path names were requested to be included.
        $thisMsg = &getMsg('msg46');
        print MISS "$thisMsg\n";
        # msg47: Source files containing only these path names were searched for:
        $thisMsg = &getMsg('msg47');
        print MISS "  $thisMsg<ul>$fontRed\n";
        for $incl (keys %includeList) {
            $excl =~ s/\\\\/\\/g;
            print MISS "$incl<br>\n";
        }
        print MISS "</font></ul><hr>\n";
        if ($onWin) {
            # Ignore case when matching pathnames on Windows.
            for $miss (sort keys %srcMissing) {
                $show = 0;
                for $incl (keys %includeList) {
                    if ($miss =~ m/$incl/i) {
                        $show = 1;
                        $miss =~ s/($incl)/$fontRed$1<\/font>/;
                        last;
                    }
                }
                print MISS $n++, ") $miss<br>\n" if $show;
            }
        } else {
            for $miss (sort keys %srcMissing) {
                $show = 0;
                for $incl (keys %includeList) {
                    if ($miss =~ m/$incl/) {
                        $show = 1;
                        $miss =~ s/($incl)/$fontRed$1<\/font>/;
                        last;
                    }
                }
                print MISS $n++, ") $miss<br>\n" if $show;
            }
        }
    } else {
        # No include or exclude list. Show all missing files.
        for $miss (sort keys %srcMissing) {
            print MISS $n++, ") $miss<br>\n";
        }
    }
    print MISS "</ul>$endHtml";
    close MISS;
    # msg25: <cmd>: Created: <file>
    &ifVerbose(__LINE__,'msg25',"$destDir$missHtml");
}

#------------------------------------------------------------------------
sub winHeader
{
    my ($fd,$level,$sp,$len) = @_;
    &setImgSrc($level,'no','blank');
    &addIt($fd,$level,$sp,$len,'show','headW',' '," |$Functions|$Functions|% $Functions|$Lines|$Lines|% $Lines");
    &setImgSrc($level,'no','blank');
    my $ucHit    = ucfirst $Hit;
    my $ucMissed = ucfirst $Missed;
    &addIt($fd,$level,$sp,$len,'show','head2',"$CovItem","$Calls|$Missed|$ucHit|$ucHit|$ucMissed|$ucHit|$ucHit");
}

#------------------------------------------------------------------------
sub unixHeader
{
    my ($fd,$level,$sp,$len) = @_;
    &setImgSrc($level,'no','blank');
    my $ucFunctions = uc $Functions;
    my $ucAdjLines  = uc $AdjLines;
    &addIt($fd,$level,$sp,$len,'show','headU',' '," | |$ucFunctions| | |$AdjLines| ");
    &setImgSrc($level,'no','blank');
    &addIt($fd,$level,$sp,$len,'show','head2',"$CovItem","$Calls|$Unused|$Used|$Used\%|$Unused|$Used|$Used\%");
}

#------------------------------------------------------------------------
sub prntWin
{
    my ($src,$object,$lnum,$line) = @_;
    my $hits = "         ";
    my $sf   = "";
    my $ef   = "";
    my $sn   = "";
    my $en   = "";
    if (defined $lineList{$src}{$lnum}) {
        my $thisHit = $lineList{$src}{$lnum};
        if ($thisHit > 0) {
            $sf   = defined $partialHit{$src}{$lnum} ? $fontMagenta : $fontBlue;
            $ef   = "</font>";
            $hits = sprintf "%9s",&insertCommas($thisHit);
        } elsif ($thisHit == 0) {
            $sf   = $fontRed;
            $ef   = "</font>";
            $hits = sprintf "%9s",$thisHit;
        }
    } elsif ($srcMap[$lnum] and $lineHits[$lnum] > 0) {
        # Mark the current line blue (same as prior line) if there is no hit count.
        $sf   = $fontBlue;
        $ef   = "</font>";
    } elsif ($srcMap[$lnum] and $lineHits[$lnum] == 0) {
        # Mark the current line red (same as prior line) if there is no hit count.
        $sf   = $fontRed;
        $ef   = "</font>";
    }
    $lnum    = sprintf "%6d",$lnum;
    $line =~ s/</\&lt;/g;
    $line =~ s/>/\&gt;/g;
    if ($object ne '') {
        my $label = "$src|$object";
        if (defined $labels{$label}) {
            $sn = "<a name=\"$object\">";
            $en = "</a>";
            delete $labels{$label};
        }
    }
    return "$sn$hits |$lnum   $sf$line$ef$en\n";
}

#------------------------------------------------------------------------
sub prntUnix
{
    my ($dir,$src,$lnum,$line) = @_;
    my $hits  = "      ";
    my $sf    = "";
    my $ef    = "";
    my $sn    = "";
    my $en    = "";
    my $Dead  = " ";
    my $Insp  = " ";
    my $Test  = " ";
    my $label = "$dir|$src|$lnum";
    my $sPath = "$dir/$src";
    if (defined $lineList{$sPath}{$lnum}) {
        my $thisHit = $lineList{$sPath}{$lnum};
        if ($thisHit > 0) {
            $sf   = $fontBlue;
            $ef   = "</font>";
            $hits = sprintf "%6d",$thisHit;
        } elsif ($thisHit == 0) {
            $sf   = $fontRed;
            $ef   = "</font>";
            $hits = sprintf "%6d",$thisHit;
        }
    } elsif ($srcMap[$lnum] and $lineHits[$lnum] > 0) {
        # Mark the current line blue (same as prior line) if there is no hit count.
        $sf   = $fontBlue;
        $ef   = "</font>";
    } elsif ($srcMap[$lnum] and $lineHits[$lnum] == 0) {
        # Mark the current line red (same as prior line) if there is no hit count.
        $sf   = $fontRed;
        $ef   = "</font>";
    }
    $lnum    = sprintf "%6d",$lnum;
    $line =~ s/</\&lt;/g;
    $line =~ s/>/\&gt;/g;
    if ($line =~ m/\/\* purecov:\s*deadcode \*\//) {
        $Dead = 'D';
    }
    if ($line =~ m/\/\* purecov:\s*inspected \*\//) {
        $Insp = "${fontBlue}I</font>";
        $sf   = $fontBlue;
        $ef   = "</font>";
    }
    if ($line =~ m/\/\* purecov:\s*tested \*\//) {
        $Test = "${fontBlue}T</font>";
        $sf   = $fontBlue;
        $ef   = "</font>";
    }
    if (defined $labels{$label}) {
        $sn = "<a name=\"$labels{$label}\">";
        $en = "</a>";
    }
    print OUTSRC "$sn$lnum |$Dead|$Insp|$Test|$sf$hits$ef | $sf$line$ef$en\n";
}

#------------------------------------------------------------------------
sub prntEpilogue
{
    my ($n,$fd)  = @_;
    my $allImg   = "";
    my $i        = 1;
    my $epilogue = $htmlEpilogue;
    if ($n == 1) {
        for (@imgList1) {
            $allImg .= "$_,";
            $allImg .= "\n" if ($i%40 == 0);
            $i++;
        }
    } else {
        for (@imgList2) {
            $allImg .= "$_,";
            $allImg .= "\n" if ($i%40 == 0);
            $i++;
        }
    }
    $allImg   =~ s/,$//;
    $epilogue =~ s/_IMAGE_MAP_/$allImg/;
    print $fd $epilogue;
}

#------------------------------------------------------------------------
sub addIt
{
    my ($fh,$level,$sp,$labelWidth,$showHide,$type,$label,$value) = @_;

    my ($calls,$funUnused,$funUsed,$funPct,$linUnused,$linUsed,$linPct) =
        split /\|/,$value;

    my ($name,$len,$left,$right,$src,$fNam,$spN,$ent,$cw,$valueU);

    $name = $label;
    $len  = length $label;
    if ($len > $labelWidth) {
        # Need to truncate label.
        if ($truncate eq 'right' or $type eq 'fun' or $type =~ m/^lnk\|fun\|/) {
            # Truncate functions/classes on the right.
            $label  = substr($label,0,$labelWidth-3) . "...";
        } elsif ($truncate eq 'center') {
            # Truncate path names (source files/directories) in the center.
            # Split in 2 parts and put dots in between them.
            $left  = int($labelWidth/2) - 1;
            $right = $left;
            $left -= 1 if (($left+$right+3) > $labelWidth);
            $left  = substr($label,0,$left);
            $right = substr($label,$len-$right,$len);
            $label = $left . "..." . $right;
        } else {
            # Truncate path name on the left.
            $label = "..." . substr($label,$len-$labelWidth+3,$len);
        }
    } elsif ($len < $labelWidth) {
        # Need to pad out the label on the right.
        $label .= " " . $spaces;
        $label  = substr $label,0,$labelWidth;
    }
    $label =~ s/</\&lt;/g;
    $label =~ s/>/\&gt;/g;
    $src   = "";
    $fNam  = "";
    if ($type =~ m/^lnk\|(\S+)\|(\S+)\|(.*)\|/) {
        # type is:  lnk|fun|srcFileName|function-name|
        #      or:  lnk|dir|dirNum|dir|
        $type = $1;
        $src  = $2;
        $fNam = $3;
    }
    $cw = 11;
    if ($type eq 'fun') {
        $funUnused = $funUnused == 1 ? $Missed : ' ';
        $funUsed   = $funUsed   == 1 ? $Hit    : ' ';
        $funUnused = sprintf "%${cw}s",$funUnused;
        $funUsed   = sprintf "%${cw}s",$funUsed;
        $funPct    = sprintf "%${cw}s"," ";  # Blank cell for functions.
    } else {
        $funUnused = sprintf "%${cw}s",$funUnused;
        $funUsed   = sprintf "%${cw}s",$funUsed;
    }
    $calls     = sprintf "%${cw}s",$calls;
    $linUnused = sprintf "%${cw}s",$linUnused;
    $linUsed   = sprintf "%${cw}s",$linUsed;
    if ($funPct =~ m/^\d*\.\d*/) {
        $funPct = sprintf "%${cw}.2f",$funPct;
        $linPct = sprintf "%${cw}.2f",$linPct;
    } else {
        $funPct = sprintf "%${cw}s",$funPct;
        $linPct = sprintf "%${cw}s",$linPct;
    }

    $value   = "$calls  $funUnused  $funUsed  $funPct  $linUnused  $linUsed  $linPct";
    if ($type eq 'fun') {
        if ($src eq '') {
            &setImgSrc($level,'no','fun');
            $calls       =~ s/^\s+//;
            $funUnused   =~ s/^\s+//;
            $funUsed     =~ s/^\s+//;
            $linUnused   =~ s/^\s+//;
            $linUsed     =~ s/^\s+//;
            $linPct      =~ s/^\s+//;
            $name        =~ s/([^\w])/\\$1/g;
            $ent         = "$sp$imgSrc <a href=\"javascript:showStats('$name',";
            $ent        .= "'$calls','$funUnused','$funUsed','$linUnused',";
            $ent        .= "'$linUsed','$linPct','$sourceFile')\">";
        } else {
            &setImgSrc($level,'popup','fun');
            $ent  = "$sp$imgSrc <a href=\"javascript:openpopup('annotated_src','$src$htmEnd#$fNam')\">";
        }
        $ent .= "$label $value</a>$br";
    } elsif ($type eq 'headU') {
        $valueU  = "$calls  ";
        $valueU .= "<u>$funUnused  $funUsed  $funPct</u>  ";
        $valueU .= "<u>$linUnused $linUsed$linPct</u>";
        $ent     = "$sp$imgSrc $label $bOn$valueU$bOff$br";
    } elsif ($type eq 'headW') {
        $ent   = "$sp$imgSrc $label $bOn$value$bOff$br";
    } elsif ($type eq 'head2') {
        # Center label.
        $label   = substr $label,0,$labelWidth;
        ($spN)   = $label =~ m/(\s+$)/;
        $label   =~ s/^/<u>/;
        $label   =~ s/:/:<\/u>/;
        $spN     = length $spN;
        $spN     = int ($spN/2);
        $spN     = " " x $spN;
        $label   =~ s/$spN$//;
        $label   = "$spN$label";
        $valueU  = "<u>$calls</u>  ";
        $valueU .= "<u>$funUnused</u>  <u>$funUsed</u>  <u>$funPct</u>  ";
        $valueU .= "<u>$linUnused</u>  <u>$linUsed</u>  <u>$linPct</u>";
        $ent     = "$sp$imgSrc $bOn$label $valueU$bOff$br";
    } else {
        if ($src eq '') {
            if ($showHide eq 'show') {
                &setImgSrc($level,'collapse',$type);
            } else {
                &setImgSrc($level,'expand',$type);
            }
            $ent  = "$sp$imgSrc <a href=\"javascript:void(0);\" onClick=\"hideShow('$imgN')\">";
            $ent .= "$label $value</a>$br";
            $ent .= "<ul class=\"$showHide\" id=\"sec$imgN\">";
        } else {
            &setImgSrc($level,'no',$type);
            my $js  = "\"javascript:openpopup('popup_dir','$src$htmEnd')\"";
            $imgSrc =~ s/>/ onClick=$js>/;
            $ent    = "$sp$imgSrc <a href=$js>";
            $ent   .= "$label $value</a>$br";
        }
    }
    print $fh $ent;
}

#------------------------------------------------------------------------
sub setImgSrc
{
    my ($level,$expCol,$type) = @_;
    my ($jPeg,$onClick);
    if ($type eq 'fun') {
        if ($expCol eq 'popup') {
            $jPeg = 'popup-fun.jpg';
        } else {
            $jPeg = 'fun.jpg';
        }
    } elsif ($type eq 'src') {
        if ($showNoSrc) {
            $jPeg = "$expCol-no-src.jpg";
        } else {
            $jPeg = "$expCol-src.jpg";
        }
    } elsif ($type eq 'popup-dir') {
        $jPeg = 'popup-dir.jpg';
    } elsif ($type eq 'dir') {
        $jPeg = "$expCol-dir.jpg";
    } elsif ($type eq 'mod') {
        $jPeg = "$expCol-mod.jpg";
    } elsif ($type eq 'expand') {
        $jPeg = 'expand.jpg';
    } elsif ($type eq 'collapse') {
        $jPeg = 'collapse.jpg';
    } elsif ($type eq 'run') {
        $jPeg = "$expCol-run.jpg";
    } elsif ($type eq 'pc') {
        $jPeg = "pc.gif";
    } elsif ($type eq 'blank') {
        $jPeg = "blank.jpg";
    } else {
        # Type should be one of: dir src mod run expand collapse
        &cmdMsg(__LINE__,'msg00',"setImgSrc(level,expCol,type) = ($level,$expCol,$type)");
        # msg33: <cmd>: ERROR: image type "<type>" not recognized.
        &cmdMsg(__LINE__,'msg33',$type);
        $imgSrc = "";
        return;
    }
    if ($level == 1) {
        $imgN1++;
        $imgN = $imgN1;
        push @imgList1,$images{$jPeg};
    } else { # level == 2
        $imgN2++;
        $imgN = $imgN2;
        push @imgList2,$images{$jPeg};
    }
    if ($expCol eq 'no' ) {
        $onClick = "";
    } else {
        $onClick = " onClick=\"hideShow('$imgN')\"";
    }
    $imgSrc  = "<img src=\"$jPeg\" name=\"img$imgN\"$onClick>";
}

#------------------------------------------------------------------------
sub setSrcMapLineHits
{
    # Mark the lines that are in a function.
    # Watch out for comments and quoting.
    my $src       = $_[0];
    @srcMap       = ();
    @lineHits     = ();
    my $lNum      = 0;
    my $inVbCode  = 0;
    my $inFunct   = 0;
    my $inComment = 0;
    my $inQuote   = 0;
    my $inQuote1  = 0;
    my $inQuote2  = 0;
    my $isVb      = $src =~ m/\.vb$/ ? 1 : 0;
    my $vbCodeTyp = $NotDefined;
    if ($isVb) {
        while (<INSRC>) {
            $lNum++;
            $srcMap[$lNum] = 0;
            if (m/(function|sub|property|operator).*\(/i) {
                $inVbCode      = 1;
                $vbCodeTyp     = $1;
                $srcMap[$lNum] = 1;
            } elsif (m/end\s+$vbCodeTyp/i) {
                $inVbCode      = 0;
                $vbCodeTyp     = $NotDefined;
                $srcMap[$lNum] = 1;
            } elsif ($inVbCode) {
                $srcMap[$lNum] = 1;
            } else {
            }
        }
    } else {
        while (<INSRC>) {
            $lNum++;
            $srcMap[$lNum] = 0;
            my $c0 = 0;
            my $c1 = 1;
            my @ln  = split //,$_;
            while ($c0 < $#ln) {
                if ($ln[$c0] eq '/' and not $inQuote) {
                    if ($ln[$c1] eq '/') {
                        last;  # The rest of the line is a comment.
                    } elsif ($ln[$c1] eq '*') {
                        # Start of a comment.
                        $inComment = 1;
                        $c0++; $c1++;
                    }
                } elsif ($ln[$c0] eq '*' and $ln[$c1] eq '/' and not $inQuote) {
                    $inComment = 0 ; $c0++; $c1++;
                } elsif ($ln[$c0] =~ m/\\/) {
                    $c0++; $c1++;  # Skip the next char.
                } elsif ($ln[$c0] eq "'" and not $inQuote2 and not $inComment) {
                    ($inQuote,$inQuote1) = ($inQuote1 and not $inQuote2) ? (0,0) : (1,1);
                } elsif ($ln[$c0] eq '"' and not $inQuote1 and not $inComment) {
                    ($inQuote,$inQuote2) = ($inQuote2 and not $inQuote1) ? (0,0) : (1,1);
                } elsif ($ln[$c0] eq '{' and not $inQuote and not $inComment) {
                    if ($inFunct) {
                        $inFunct++;
                    } else {
                        $inFunct = 1;
                        $srcMap[$lNum] = 1;
                    }
                } elsif ($ln[$c0] eq '}' and $inFunct and not $inQuote and not $inComment) {
                    $inFunct--;
                    if ($inFunct == 0) {
                        $srcMap[$lNum] = 1;
                    }
                }
                $c0++; $c1++;
            }
            if ($inFunct) {
                $srcMap[$lNum] = $inFunct;
            }
        }
    }
    seek INSRC,0,0;
    $priorHit = -1;
    for ($lNum=$#srcMap+1; $lNum>0; $lNum--) {
        if (defined $lineList{$src}{$lNum}) {
            $priorHit = $lineList{$src}{$lNum};
            $lineHits[$lNum] = $priorHit;
        } elsif ($srcMap[$lNum]) {
            $lineHits[$lNum] = $priorHit;
        } else {
            $lineHits[$lNum] = -1;
        }
    }

}

#------------------------------------------------------------------------
sub totalDir
{
    # Keep a running total of each directory for the directory view.
    my ($dir,$entry) = @_;
    if (not defined $dir2List{$name}) {
        return $entry;
    }
    my ($calls,$funUnused,$funUsed,$funPct,$linUnused,$linUsed,$linPct) =
        split /\|/,$entry;
    my ($calls2,$funUnused2,$funUsed2,$funPct2,$linUnused2,$linUsed2,$linPct2) =
        split /\|/,$dir2List{$name};
    $calls     += $calls2;
    $funUnused += $funUnused2;
    $funUsed   += $funUsed2;
    $linUnused += $linUnused2;
    $linUsed   += $linUsed2;
    my $funTot  = $funUsed + $funUnused;
    my $linTot  = $linUsed + $linUnused;
    if ($funTot) {
        $funPct = sprintf "%.2f",$funUsed/$funTot*100;
    } else {
        $funPct = '0.00';
    }
    if ($linTot) {
        $linPct = sprintf "%.2f",$linUsed/$linTot*100;
    } else {
        $linPct = '0.00';
    }
    return "$calls|$funUnused|$funUsed|$funPct|$linUnused|$linUsed|$linPct";
}

#------------------------------------------------------------------------
sub determineOs
{
    # Return onUnix,onWin:
    #   onUnix=0, onWin=1 if on Windows ($^O = MSWin32 or MSWinNT).
    #   onUnix=1, onWin=0 if on UNIX (uname command exists).
    if ($^O =~ m/MSWin/) {
        return (0,1);
    } elsif (-e "/bin/uname" or -e "/usr/bin/uname")  {
        return (1,0);
    } else {
        # msg34: <cmd>: You must run <cmd> on UNIX or Windows.
        &cmdMsg(__LINE__,'msg34',$cmdName);
        # msg35: <cmd>: $^O is "<value-of-$^O>", not "MSWin" and [/usr]/bin/uname not found.
        &cmdMsg(__LINE__,'msg35',$^O,'[/usr]/bin/uname');
        &doExit(1);
    }
}

#------------------------------------------------------------------------
sub insertCommas
{
    my $number = reverse $_[0];
    $number =~ s/(\d\d\d)(?=\d)(?!\d*\.)/$1,/g;
    return scalar reverse $number;
}

#------------------------------------------------------------------------
#YYsub checkModTime
#{
#    # If the file (arg1) modification time is greater than the start of the
#    # PureCoverage data collection run (epochStart), return true: warn user.
#    my ($file) = @_;
#    return 0 unless ($epochStart);  # Cannot check if epochStart not set.
#    my $mTime  = (stat($file))[9];
#    if ($mTime > $epochStart) {
#        # msg37: <cmd>: Warning: <file>
#        &cmdMsg(__LINE__,'msg37',$file);
#        # msg36: <cmd>: This source file was modified since the start of the PureCoverage data collection
#        &cmdMsg(__LINE__,'msg36');
#        return 1;
#    } else {
#        return 0;
#    }
#}
#
#------------------------------------------------------------------------
#YYsub getEpoch
#{
#    my ($sec,$min,$hour,$mday,$mon,$year) = @_;
#    my %months = ('jan',1,'feb',2,'mar',3,'apr',4,'may',5,'jun',6,
#                  'jul',7,'aug',8,'sep',9,'oct',10,'nov',11,'dec',12);
#    use Time::Local;
#    if ($mon =~ m/\w{3}/) {
#        $mon = lc $mon;
#        $mon = $months{$mon}-1;
#    } else {
#        $mon =~ s/^0//;
#        $mon--;
#    }
#    return timelocal($sec,$min,$hour,$mday,$mon,$year-1900);
#}

#------------------------------------------------------------------------
sub getKeys
{
    my ($keyList,$mod,$dir,$src) = @_;
    if ($keyList eq 'modList') {
        if ($mod) {
            return sort keys %modList;
        } else {
            return 0;
        }
    }
    elsif ($keyList eq 'dirList') {
        if ($mod) {
            return sort keys %{$dirList{$mod}};
        } else {
            return sort keys %dir2List;
        }
    }
    elsif ($keyList eq 'srcList') {
        if ($mod) {
            return sort keys %{$srcList{$mod}{$dir}};
        } else {
            return sort keys %{$src2List{$dir}};
        }
    }
    elsif ($keyList eq 'funList') {
        if ($mod) {
            return sort keys %{$funList{$mod}{$dir}{$src}};
        } else {
            return sort keys %{$fun2List{$dir}{$src}};
        }
    }
}

#------------------------------------------------------------------------
sub getObject
{
    my $object = $_[0] . '_';
    my $method = $_[1];
    $method    =~ s/~/tld/g;
    $method    =~ s/!/exc/g;
    $object   .= $method;
    $object    =~ s/\W//g;
    return $object;
}

#------------------------------------------------------------------------
sub getHostName
{
    if ($onWin) {
        return (gethostbyname('localhost'))[0];
    } else {
        # on UNIX use uname command.
        my $host = `uname -n`; chomp $host;
        return $host;
    }
}

#------------------------------------------------------------------------
sub getUserName
{
    if ($onWin) {
        return getlogin || (getpwuid($<))[0] || "guest";
    } else {
        my $def = defined $ENV{USER} ? $ENV{USER} : "guest";
        return getlogin || (getpwuid($<))[0] || $def;
    }
}

#------------------------------------------------------------------------
sub getPaths
{
    my ($path,@path1);
    if ($onWin) {
        # Return each item in the path using the ; separator.
        @path1 = split /;/,$_[0];
    } else {
        # Return each item in the path using the : separator.
        @path1 = split /:/,$_[0];
    }
    # Check for ENV_VAR's and expand them first.
    my @path2 = ();
    for $path (@path1) {
        if ($path =~ m/^[A-Z0-9_]+$/) {
            if (defined $ENV{$path}) {
                if ($ENV{$path} ne '') {
                    # msg38: <cmd>: Using <path> to find source files: $ENV{$path}
                    &ifVerbose(__LINE__,'msg38',$path,$ENV{$path});
                    if ($onWin) {
                        push @path2, split /;/,$ENV{$path};
                    } else {
                        push @path2, split /:/,$ENV{$path};
                    }
                } else {
                    # msg39: <cmd>: <path> not set.
                    &ifVerbose(__LINE__,'msg39',$path);
                }
            } else {
                # msg40: <cmd>: <path> not defined.
                &ifVerbose(__LINE__,'msg40',$path);
            }
        } else {
            push @path2,$path;
        }
    }
    my @path3 = ();
    for $path (@path2) {
        if ($path =~ m/(.*)[\/\\]\*\*$/) {
            # Do a recursive search for sub-directories and folders.
            if (-d $1) {
                &recursePath(\@path3,$1);
            } else {
                # msg41: <cmd>: <dir> not found, not a Directory.
                &cmdMsg(__LINE__,'msg41',$1,$Directory);
            }
        } elsif (-d $path) {
            push @path3,$path;
        } else {
            $path =~ s/[\/\\][^\/\\]+$//;
            push @path3,$path;
        }
    }
    return @path3;
}

#------------------------------------------------------------------------
sub splitPaths
{
    my ($path,@path1);
    if ($onWin) {
        # Return each item in the path using the ; separator.
        @path1 = split /;/,$_[0];
    } else {
        # Return each item in the path using the : separator.
        @path1 = split /:/,$_[0];
    }
    return @path1;
}

#------------------------------------------------------------------------
sub recursePath
{
    my ($pathRef,$path) = @_;
    push @$pathRef,$path;
    my $item;
    if (opendir(PATH,$path)) {
        # Search dir, ignore . (current dir) and .. (parent dir).
        my @items = grep !/^\.\.?\z/, readdir PATH;
        close PATH;
        for $item (@items) {
            if (-d "$path$sep$item") {
                &recursePath($pathRef,"$path$sep$item");
            }
        }
    }
}

#------------------------------------------------------------------------
sub loadMsgVar
{
    unless (open(MSG,"<$msgFile")) {
        # Cannot use the message catalog.  Print this one in English only.
        &cmdMsg(__LINE__,'msg00',"Failed to open $msgFile for reading: $!");
        unless (-d $imgDir) {
            my $folder = $onWin ? 'folder' : 'directory';
            &cmdMsg(__LINE__,'msg00',"The supporting files $folder does not exist:");
            &cmdMsg(__LINE__,'msg00',"  $imgDir");
        }
        &doExit(1);
    }
    unless (-d $imgDir) {
        &myDie(__LINE__,'msg08',$Directory,$imgDir);
    }
    my $ctrlM = chr(13);
    while (<MSG>) {
        s/$ctrlM$//;  # Remove any ctrl-M chars terminating the line.
        if (m/^(var\d+)\s+\"(.*)\"$/) {
            $var{$1} = $2;
        } elsif (m/^(msg\d+)\s+\"(.*)\"$/) {
            $msgCat{$1} = $2;
        }
    }
    close MSG;
}

#------------------------------------------------------------------------
sub usage
{
    my $lin = shift;
    my $msg = shift;
    if (defined $_[0]) {
        &cmdMsg($lin,$msg,@_);
        # msg42: <cmd>: Run "<cmd> -help" for usage information.
        &cmdMsg(__LINE__,'msg42',$cmdName);
        &doExit(1);
    }
    my $indent  = 0;
    my $indLvl  = 5;
    my $center  = 0;
    my $buffer  = "";
    my $under   = "";
    my $bold    = "";
    unless (open(FILE,"<$cmdHelp")) {
        # msg30: <cmd>: Cannot read <file>: <status>
        &myDie(__LINE__,'msg30',$cmdHelp,$!);
    }
    open(MORE,"|more");
    # msg43: <cmd>: Converting <file> to text...
    $thisMsg = &getMsg('msg43',$cmdHelp);
    print MORE "$cmdName $version $rev: $thisMsg\n";
    while (<FILE>) {
        last if (m/^<body /);
    }
    while (<FILE>) {
        chomp;
        next if (m/^<tr>/i);
        s/<i>/\</gi;
        s/<\/i>/\>/gi;
        s/<b>//g;
        s/<\/b>//g;
        s/<\/font>//gi;
        s/<font face=\w+>//gi;
        s/\&nbsp; //gi;
        if (m/<br>/i) {
            s/<br>//gi;
            $buffer .= " $_";
            $buffer  = &helpOut($indent,$center,$buffer);
        } elsif (m/^<ul>$/i){
            $buffer = &helpOut($indent,$center,$buffer);
            print MORE "\n";
            $indent += $indLvl;
        } elsif (m/^<\/ul>$/i) {
            $buffer = &helpOut($indent,$center,$buffer);
            $indent -= $indLvl;
        } elsif (m/^<center>$/i){
            $buffer = &helpOut($indent,$center,$buffer);
            print MORE "\n";
            $center = 1;
        } elsif (m/^<\/center>$/i){
            $buffer = &helpOut($indent,$center,$buffer);
            $center = 0;
        } elsif (m/^<\/html>$/i){
            $buffer = &helpOut($indent,$center,$buffer);
        } elsif (m/^<p>$/i){
            $buffer = &helpOut($indent,$center,$buffer);
            print MORE "\n";
        } elsif (m/^<B>(.*)<\/B>$/i){
            $bold   = uc $1;
            $buffer = &helpOut($indent,$center,$buffer);
            $buffer = &helpOut($indent,$center,$bold);
        } elsif (m/^<u>(.*)<\/u>$/i){
            $under  = $1;
            $buffer = &helpOut($indent,$center,$buffer);
            $buffer = &helpOut($indent,$center,$under);
            $under  =~ s/./-/g;
            $buffer = &helpOut($indent,$center,$under);
        } elsif ($_ !~ m/^</){
            $buffer .= " $_";
        }
    }
    close FILE;
    if ($onWin) {
        if ($imgDir =~ m/^pure/) {
            $imgDir = ".\\$imgDir";
        } else {
            $imgDir =~ s/\//\\/g;
        }
    } else {
        if ($imgDir =~ m/^pure/) {
            $imgDir = "./$imgDir";
        }
    }
    if (-d $imgDir) {
        $buffer = &helpOut($indLvl,0,"The JPEG images are in this $Directory");
    } else {
        $buffer = &helpOut($indLvl,0,"The JPEG images $Directory was not found:");
    }
    $buffer = &helpOut($indLvl*2,0,$imgDir);
    close MORE;
    &doExit(0);
}

#------------------------------------------------------------------------
sub helpOut
{
    my ($indent,$center,$out) = @_;
    return "" if ($out eq '');
    return "" if ($out =~ m/^\s+$/);
    my ($thisInd,$maxLine,$thisLen,$word,$lineOut);
    if ($center) {
        $indent  = 80 - length $out;
        $thisInd = " " x int($indent/2 - 1);
        print MORE "$thisInd$out\n";
        return "";
    }
    # Split $out into separate lines that are no longer than 78 chars, including the indent
    $thisInd = " " x $indent;
    $maxLine = 78;
    $lineOut = $thisInd;
    for $word (split / /,$out) {
        next if ($word eq '');
        $thisLen = length "$lineOut$word";
        if ($thisLen >= $maxLine) {
            print MORE "$lineOut\n";
            $lineOut  = "$thisInd$word";
        } else {
            $lineOut .= $word;
        }
        $lineOut .= ($word =~ m/\.$/) ? '  ' : ' ';
    }
    print MORE "$lineOut\n";
    return "";
}

#------------------------------------------------------------------------
sub setCharSetBold
{
    ($charSet,$boldIs) = @_;
    if ($boldIs eq 'on') {
        # Turn on bold .
        $bOn      = "<b>";  # Turn bold on
        $bOff     = "</b>"; # Turn bold off
        $fontBold = "; font-weight:bold";
    } else {
        # Turn off bold .
        $bOn      = '';
        $bOff     = '';
        $fontBold = '';
    }
}

#------------------------------------------------------------------------
sub ifVerbose
{
    return unless $verbose;
    &cmdMsg(@_);
}

#------------------------------------------------------------------------
sub myDie
{
    &cmdMsg(@_);
    &doExit(1);
}

#------------------------------------------------------------------------
sub cnvtMsg
{
    my $msg = shift;
    $msg = $msgCat{$msg};
    my $i = 1;
    for (@_) {
        $msg =~ s/\%s$i/$_/;
        $i++;
    }
    return $msg;
}

#------------------------------------------------------------------------
sub getMsg
{
    my $msg = shift;
    if (@_) {
        return &cnvtMsg($msg,@_);
    } else {
        return $msgCat{$msg};
    }
}

#------------------------------------------------------------------------
sub cmdMsg
{
    my $lin = shift;
    my $msg = shift;
    my $msn = $msg; $msn =~ s/msg//;
    if ($msg eq 'msg00') {
        $msg = $_[0];
    } elsif (@_) {
        $msg = &cnvtMsg($msg,@_);
    } else {
        $msg = $msgCat{$msg};
    }
    my $out  = sprintf "%s:%04d-%02d: %s\n",$cmdName,$lin,$msn,$msg;
    print $out unless ($cmdOutOnly);
    $cmdOut .= $out;
    $cmdOutN++;
}

__END__

