using System;
using System.ServiceProcess;
using System.Collections.Generic;
using System.Collections;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Security.Permissions;
using System.Security.AccessControl;
using System.Configuration.Install;
using System.Windows.Forms;
using System.Resources;

namespace ManagedServiceClient
{
    public partial class ClientForm : Form
    {
        private Boolean bServiceInstalled = false;
        private string strServiceName = "ManagedService";
        private string strAssemblyName = "ManagedServiceSample.exe";
        ServiceControllerPermission permission;
        ManagedServiceSample.RemoteService remote;
        private string strRemote = "tcp://localhost:8085/ibm.rational.purifyplus.managedservicesample";

        public ClientForm()
        {
            InitializeComponent();
            
            
        }

        private void ClientForm_Load(object sender, EventArgs e)
        {
            try
            {
                bServiceInstalled = FindService(strServiceName);
                if (bServiceInstalled)
                {
                    permission = new ServiceControllerPermission(ServiceControllerPermissionAccess.Control, ".", strServiceName);
                    permission.Demand();
                    remote = (ManagedServiceSample.RemoteService)
                        Activator.GetObject(typeof(ManagedServiceSample.RemoteService), strRemote);
                    if (remote == null)
                    {
                        textBoxService.Text = "Communication problem";
                    }
                    else
                    {
                        textBoxService.Text = remote.Status();
                    }
                }
            }
            catch (Exception ex)
            {
                string s = ex.ToString();
                toolStripStatusLabel.Text = ex.Message;
            }
            RecalcLayout();
        }

        private bool FindService(string i_strName)
        {
            Boolean bRet = false;
            ServiceController[] slist = ServiceController.GetServices();
            foreach (ServiceController sc in slist)
            {
                if (sc.ServiceName == i_strName)
                {
                    bRet = true;
                }
            }
            return bRet;
        }

        private void RecalcLayout()
        {
            if (bServiceInstalled)
            {
                buttonInstall.Enabled = false;
                buttonRemove.Enabled = true;
                serviceController.ServiceName = strServiceName;
                switch (serviceController.Status)
                {
                    case ServiceControllerStatus.Stopped:
                        toolStripStatusLabel.Text = "Stopped";
                        buttonStart.Enabled = true;
                        buttonStop.Enabled = false;
                        buttonPerfTest.Enabled = false;
                        buttonLeakTest.Enabled = false;
                        break;
                    case ServiceControllerStatus.Running:
                        toolStripStatusLabel.Text = "Running";
                        buttonStart.Enabled = false;
                        buttonStop.Enabled = true;
                        buttonPerfTest.Enabled = true;
                        buttonLeakTest.Enabled = true;
                        break;
                    default:
                        break;
                }
            }
            else
            {
                buttonInstall.Enabled = true;
                buttonRemove.Enabled = false;
                toolStripStatusLabel.Text = "Managed Service is not installed";
            }
        }

        private void buttonStart_Click(object sender, EventArgs e)
        {
            toolStripStatusLabel.Text = "Starting...";
            serviceController.Start();
            serviceController.WaitForStatus(ServiceControllerStatus.Running);
            RecalcLayout();
        }

        private void buttonStop_Click(object sender, EventArgs e)
        {
            toolStripStatusLabel.Text = "Stopping...";
            serviceController.Stop();
            serviceController.WaitForStatus(ServiceControllerStatus.Stopped);
            RecalcLayout();
        }

        private void buttonInstall_Click(object sender, EventArgs e)
        {
            // Install managed service
            IDictionary savedState = new Hashtable();
            string[] commandLineOptions = new string[1] { "/LogFile=example.log" };
            try
            {
                AssemblyInstaller myAssemblyInstaller = new 
                         AssemblyInstaller( strAssemblyName, commandLineOptions );
                myAssemblyInstaller.UseNewContext = true;
                myAssemblyInstaller.Install( savedState );
                myAssemblyInstaller.Commit( savedState );
                bServiceInstalled = true;
                RecalcLayout();
            }
            catch (ArgumentException)
            {
                toolStripStatusLabel.Text = "Install failed";
            }
            catch (Exception ex)
            {
                toolStripStatusLabel.Text = ex.Message;
            }
           
        }

        private void buttonRemove_Click(object sender, EventArgs e)
        {
            IDictionary savedState = new Hashtable();
            try
            {
                AssemblyInstaller myAssemblyInstaller = new
                         AssemblyInstaller(strAssemblyName, new String[0]);
                myAssemblyInstaller.Uninstall(savedState);
                bServiceInstalled = false;
                RecalcLayout();
            }
            catch (ArgumentException aex)
            {
                toolStripStatusLabel.Text = aex.Message;
            }
            catch (Exception ex)
            {
                toolStripStatusLabel.Text = ex.Message;
            }
        }

        private void buttonPerfTest_Click(object sender, EventArgs e)
        {
            toolStripStatusLabel.Text = "";
            try
            {
                remote.TestPerformance(3, 1000);
                textBoxService.Text = remote.Status();
            }
            catch (Exception ex)
            {
                toolStripStatusLabel.Text = ex.Message;
            }
        }

        private void buttonLeakTest_Click(object sender, EventArgs e)
        {
            toolStripStatusLabel.Text = "";
            try
            {
                remote.TestMemory(10, 1000);
                textBoxService.Text = remote.Status();
            }
            catch (Exception ex)
            {
                toolStripStatusLabel.Text = ex.Message;
            }
        }
    }
}