//==============================================================================
// DotNetClientW::SampleServiceFacade
//
// Sample Service Facade Interface
// Used by DotNetClientW to controll and interract with the sample service
// Derived from Service Controller
//==============================================================================
#pragma once

#pragma region using
using namespace SampleServiceLib;
using namespace System;
using namespace System::ServiceProcess;
#pragma endregion

namespace DotNetClientW
{
	public ref class SampleServiceFacade : public ServiceController
	{
	//--------------------------------------------------------------------------
	// Constructor
	// Attempts to create ServiceController for "SampleService
	// May throw an exception
	//
	public: SampleServiceFacade() 
				: m_strName("SampleService")
				, ServiceController(m_strName)
			{
				m_irClient = nullptr;
				ServiceController(m_strName);
				Refresh();
				if(Status == ServiceControllerStatus::Running)
					m_irClient = gcnew ClientInterfaceClass();
			}

	//--------------------------------------------------------------------------
	// Destructor
	//
	public: ~SampleServiceFacade()
			{
				delete m_irClient;
				Close();
			}

	//--------------------------------------------------------------------------
	// Start Service
	// In: i_uTimeoutMillis - Service timeout interval in milliseconds
	// Note: an exception will be thrown if service doesn't start within specified timeout
	//
	public: System::Void Start(unsigned i_uTimeoutMillis) 
			{
				__super::Start();
				__super::WaitForStatus(ServiceControllerStatus::Running, TimeSpan(i_uTimeoutMillis * 10000));
				m_irClient = gcnew ClientInterfaceClass();
			}

	//--------------------------------------------------------------------------
	// Stop Service
	// In: i_uTimeoutMillis - Service timeout interval in milliseconds
	// Note: an exception will be thrown if service doesn't stop within specified timeout
	//
	public: System::Void Stop(unsigned i_uTimeoutMillis)
			{
				__super::Stop();
				__super::WaitForStatus(ServiceControllerStatus::Stopped, TimeSpan(i_uTimeoutMillis * 10000));
				m_irClient = nullptr;
			}

	//--------------------------------------------------------------------------
	// Property: StatusString (Read-Only)
	// This property contains a string corresponding to the service status.
	//
	public: property String^ StatusString
			{
				String^ get()
				{
					Refresh();
					return Status.ToString();
				}
			}
	
	//--------------------------------------------------------------------------
	// DoUMR
	// Trigger UMR in the instrumented code.
	//
	public: System::Void DoUMR()
			{
				if(m_irClient != nullptr)
				{
					m_irClient->Control(enControl::CONTROL_UMR);
				}
			}
	
	//--------------------------------------------------------------------------
	// DoABR
	// Trigger ABR in the instrumented code
	//
	public: System::Void DoABR()
			{
				if(m_irClient != nullptr)
				{
					m_irClient->Control(enControl::CONTROL_ABR);
				}
			}

	//--------------------------------------------------------------------------
	// DoMLK
	// Trigger MLK in th einstrumented code
	//
	public: System::Void DoMLK()
			{
				if(m_irClient != nullptr)
				{
					m_irClient->Control(enControl::CONTROL_MLK);
				}
			}

	//--------------------------------------------------------------------------
	// GetEnv
	// Get the environment variable
	//		Input: i_strName -- Environment variable name
	//		Returns value of the requested variable 
	// Note: the variable will be retrieved from the service
	//		environment.
	//
	public: String^ GetEnv(String ^i_strName)
			{
				String^ strVal = m_irClient->GetEnv(i_strName);
				return strVal;
			}

	//--------------------------------------------------------------------------
	// IsInstrumented
	// Retruns instrumentation type for the service or
	// INSTRUMENTED_NOTHING if it is not instrumented.
	//
	public: enum enInstrumented IsInstrumented()
			{
				enum enInstrumented enRet = enInstrumented::INSTRUMENTED_NOTHING;
				if(m_irClient != nullptr)
				{
					enRet = m_irClient->IsInstrumented();
				}
				return enRet;
			}

	//--------------------------------------------------------------------------
	// Private members
	private: String^ m_strName;					// Service Name
	private: ClientInterfaceClass^ m_irClient;	// Sample Service Client Interface

	};

}

//= End of SampleServiceFacade.h ===============================================