//==============================================================================
// ClientInterface.cpp 
// Implementation of CClientInterface
//
// DCOM ProgID: SampleService.ClientInterface
// CoClass GUID: {0293D365-1D51-450E-9CBB-89A2CA541B77}
// IID_IClientInterface: {8B705020-6322-4466-81F1-78502D944DA4}
// TypeLib: {925DF074-A317-4F39-AC00-4AD9FD852090}
//
//==============================================================================

#include "stdafx.h"
#include "ClientInterface.h"
// Pure_api interface. See PurifyPlus doccumentation for details.
extern "C"
{
#include "pure_api.c"
}
//------------------------------------------------------------------------------
// Public implementation


//------------------------------------------------------------------------------
// Service control function.
// Performes an action based on the control code.

STDMETHODIMP CClientInterface::Control(enControl_t i_ulCode)
{

	HRESULT hr = S_OK;
	switch(i_ulCode)
	{
	case CONTROL_ABR:
		DoABR();
		break;
	case CONTROL_UMR:
		DoUMR();
		break;
	case CONTROL_MLK:
		DoMLK();
		break;
	default:
		break;
	}
	return hr;
}

//------------------------------------------------------------------------------
// Returns the requested environment variable from system to client
//
STDMETHODIMP CClientInterface::GetEnv(BSTR i_bstrName, BSTR* r_pbstrVal)
{
	CString strName(i_bstrName); 
	//wchar_t* wszVal = _wgetenv((LPCTSTR)strName);
	wchar_t* wszVal = NULL;
	errno_t err = _wdupenv_s(&wszVal, NULL, (LPCWSTR)strName);
	if(0 == err)
	{
		*r_pbstrVal = ::SysAllocString(wszVal);
		free(wszVal);
	}
	else
	{
		*r_pbstrVal = NULL;
	}
	return S_OK;
}

//------------------------------------------------------------------------------
// Returns the requested environment variable from system to client
//
// Returns:
//		INSTRUMENTED_NOTHING  = 0  If not instrumented
//		INSTRUMENTED_PURIFY   = 1  for Purify
//		INSTRUMENTED_COVERAGE =	2  for Coverage
//		INSTRUMENTED_QUANTIFY = 3  for Quantify
//
STDMETHODIMP CClientInterface::IsInstrumented(enInstrumented_t* r_enInstrumented)
{
	enum enInstrumented enResult = INSTRUMENTED_NOTHING;
	if(PurifyIsRunning())
	{
		 enResult = INSTRUMENTED_PURIFY;
	}
	else if(CoverageIsRunning())
	{
		enResult = INSTRUMENTED_COVERAGE;
	}
	else if(QuantifyIsRunning())
	{
		enResult = INSTRUMENTED_QUANTIFY;
	}
	*r_enInstrumented = enResult;
	return S_OK;
}

//------------------------------------------------------------------------------
// Private implementation


//------------------------------------------------------------------------------
// Generates an ABR
void
CClientInterface::DoABR()
{
	const size_t sizeForABR = 12;
	TCHAR* szForABR = new TCHAR[sizeForABR];
	TCHAR tc = szForABR[sizeForABR + 1];
	delete szForABR;
}

//------------------------------------------------------------------------------
// Generates an UMR
void
CClientInterface::DoUMR()
{
	unsigned* au = new unsigned[10];
	bool random = false;
	if(au[0])
	{
		random = true;
	}
	delete au;
}

//------------------------------------------------------------------------------
// Leaks 128 bytes of memory
static void InternalDoMLK()
{
	unsigned char* pbBuffer = new unsigned char[128];
}

void 
CClientInterface::DoMLK()
{
	InternalDoMLK();
	PurifyNewLeaks();
}

//= End of ClientInterface.cpp =================================================
