//==============================================================================
// SimpleClient.cpp 
// Simple client of the SampleService
//
// SampleService has to be installed prior to running this test.
// If the service is instrumented by purify, the run should result in 
// one UMR, one ABR and one MLK.
// To end purify session, stop the service in the service manager
// or use "net stop SampleService" command.
//
// Return codes:
//		0 - Success
//		1 - DCOM Error
//		2 - System error
//
// --vlh--
//==============================================================================

#include "stdafx.h"
using namespace ATL;
using namespace std;

// Return codes
typedef enum
{
	RET_OK,
	RET_COM,
	RET_FATAL,
} retcode_t;

// COM Interface for the SampleService client
typedef CComPtr<IClientInterface> comptrClient_t;

// TestService
void TestSampleService()
{
		comptrClient_t comptrClient;
		HRESULT hr = S_OK;
		// Default COM initialization
		hr = CoInitialize(NULL);	
		// Get instance of IClientInterface
		hr = comptrClient.CoCreateInstance(OLESTR("SampleService.ClientInterface"), 0, CLSCTX_LOCAL_SERVER | CLSCTX_REMOTE_SERVER);
		if(hr) throw hr;
		IClientInterface* piClient = (IClientInterface*)comptrClient;
		// Check the globabl namespace setting
		LPCTSTR szEnvGlobalNamespaceName = _T("PURE_USE_GLOBAL_NAMESPACE");
		BSTR bstrEnvGlobalNamespace = piClient->GetEnv(szEnvGlobalNamespaceName);
		if(NULL == bstrEnvGlobalNamespace)
		{
			tcout << _T("Environment variable ") << szEnvGlobalNamespaceName << _T(" is not set.") << endl;
			tcout << _T("Remember to add it to the SYSTEM environment when using a remote session.") << endl;
		}
		else
		{
			CString strVal = bstrEnvGlobalNamespace;
			tcout << szEnvGlobalNamespaceName << _T(" = ") << (LPCTSTR)strVal << endl;
			::SysFreeString(bstrEnvGlobalNamespace);
		}
		// Try all control codes
		tcout << "Service is ";
		if(piClient->IsInstrumented())
		{
			for(int nCode = CONTROL_NOTHING; nCode <= CONTROL_MLK; nCode++)
			{
				hr = piClient->Control((enControl)nCode);
				if(hr) throw hr;
			}
		}
		else 
		{
			tcout << "not ";
		}
		tcout << "instrumented." << endl;
		tcout << _T("SampleService test complete.") << endl;
		tcout << _T("Stop the SampleService to end PurifyPlus session.") << endl;
}

// main
int _tmain(int argc, _TCHAR* argv[])
{
	retcode_t retcode = RET_OK;
	try
	{
		TestSampleService();
	}
	catch(HRESULT hrEx)
	{	// COM exception
		tcerr << _T("COM Error ") << hex << hrEx << endl;
		retcode = RET_COM;
	}
	catch(...)
	{	// Other exception
		DWORD dwErr = ::GetLastError();
		if(dwErr)
		{
			tcerr << _T("System Error ") << hex << dwErr << endl;
		}
		else
		{
			tcerr << _T("Unknown Error") << endl;
		}
		retcode = RET_FATAL;
	}
	return retcode;
}

//= End of SimpleClient.cpp ====================================================
