//*
//* stock.h
//*
//* | Licensed Materials - Use restricted, please refer to the "Samples" terms
//* | and conditions in the IBM International Program License Agreement (IPLA).  
//* (C) Copyright IBM Corporation. 1992, 2006. All Rights Reserved.
//*

// Stock.h : main header file for the STOCK application
//

#ifndef __AFXWIN_H__
	#error include 'stdafx.h' before including this file for PCH
#endif

#include "resource.h"       // main symbols

// function prototype that converts scaled price integer to a string.
char* GetPriceStr (int, char*);

/////////////////////////////////////////////////////////////////////////////
// CSymArray this class contains the string symbols of all stocks traded.
// It is used to generate the random stock transactions that drive the
// stock simulation.

class CSymbolArray : public CObject
{
public:
	int m_LastSymbol;
	CSymbolArray ();
	int GenTransaction ();	//generates a random stock transaction
	int GenDelta (int PIndex);	//generates stock price change value
	int GetIndx (char* PSymName);	//returns index based on symbol name
	char* GetName (int PIndx);		//finds the symbol name for indx
	int GetPrice (int PIndx);		//returns the price for symbol indx
	void Restart ();	//restarts GetNext
	char* GetNext ();	//returns next symbol in list not already displayed
	char* GetNextAll ();			//returns next symbol in list
	void SetDoc (int PIndx, CDocument* PDoc);	//sets the document handler
};
extern CSymbolArray StockSymbols;	//only instance declared 

/////////////////////////////////////////////////////////////////////////////
// CStockApp:
// See Stock.cpp for the implementation of this class
//

#define TICKER_SIZE	80	// size of string holding ticker tape
#define SYMBOL_SIZE	 8	// size of a stock symbol
#define PRICE_SIZE	 8	// width of the price, including fractions
#define SCROLL_OFFSET	16	// how much to move the tape along each tick
#define NEW_SYM_START	(TICKER_SIZE - SCROLL_OFFSET)
#define NEW_PRC_START	(TICKER_SIZE - SCROLL_OFFSET) + SYMBOL_SIZE

class CStockApp : public CWinApp
{
public:
	CMultiDocTemplate* m_pStockDocTemplate;
	CMultiDocTemplate* m_pSymDocTemplate;
	CStatusBar* m_wndStatusBar;
	int   m_TickerCounter;
        char* m_Ticker;
	CStockApp();
	void CStockApp::MoveTicker();	//moves the ticker tape 

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CStockApp)
	public:
	virtual BOOL InitInstance();
	//}}AFX_VIRTUAL

// Implementation

	//{{AFX_MSG(CStockApp)
	afx_msg void OnAppAbout();
	afx_msg void OnFileNew();
	afx_msg void OnViewStocks();
	afx_msg void OnViewOneStock();
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

extern CStockApp theApp;	//the only application instance

/////////////////////////////////////////////////////////////////////////////
