//*
//* stockdoc.cpp
//*
//* | Licensed Materials - Use restricted, please refer to the "Samples" terms
//* | and conditions in the IBM International Program License Agreement (IPLA).  
//* (C) Copyright IBM Corporation. 1992, 2006. All Rights Reserved.
//*

// Stockdoc.cpp : implementation of the CStockDoc class
//

#include "stdafx.h"
#include "Stock.h"

#include "Stockdoc.h"
#include "SellDlg.h"
#include "BuyDlg.h"
#include "ShortDlg.h"

#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CStockDoc

IMPLEMENT_DYNCREATE(CStockDoc, CDocument)

BEGIN_MESSAGE_MAP(CStockDoc, CDocument)
	//{{AFX_MSG_MAP(CStockDoc)
	ON_COMMAND(ID_TRADE_BUY, OnTradeBuy)
	ON_COMMAND(ID_TRADE_SELL, OnTradeSell)
	ON_COMMAND(ID_TRADE_SHORT, OnTradeShort)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CStockDoc construction/destruction

CStockDoc::CStockDoc()
{
}

/////////////////////////////////////////////////////////////////////////////
// CStockDoc diagnostics

#ifdef _DEBUG
void CStockDoc::AssertValid() const
{
	CDocument::AssertValid();
}

void CStockDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CStockDoc commands

CPosition::CPosition () : m_strDisplay ()
{
}

CPosition::CPosition (CPosition& PPos) : m_strDisplay (PPos.m_strDisplay)
{
	m_Symbol = PPos.m_Symbol;
	m_Price = PPos.m_Price;
	m_Shares = PPos.m_Shares;
}

CPosition::CPosition (int PSymbol, int PPrice, int PShares) 
	: m_strDisplay ()
{
	m_Symbol =PSymbol; 
	m_Shares = PShares; 
	SetPrice (PPrice);
}

void CPosition::SetPrice (int PPrice)
{
	char strTemp [20];
	m_Price = PPrice;
	m_strDisplay.Format ("%s  %s  %4.0i @ $%s", StockSymbols.GetName (m_Symbol), 
		(m_Shares < 0)? "short" : "long", m_Shares, 
		GetPriceStr (PPrice, strTemp) );
}

CPosition& CPosition::operator= (CPosition& rVal)
{
	m_Symbol = rVal.m_Symbol;
	m_Price = rVal.m_Price;
	m_Shares = rVal.m_Shares;
	m_strDisplay = rVal.m_strDisplay;
	return *this;
}

//CStockDoc::OnTradeBuy - opens a dialog and if the user selects a
// stock and amount it adds a position to this document. Then it calls
// update all views to display the new position.
void CStockDoc::OnTradeBuy() 
{
	int i;
	CBuyDlg dlg;
	char *strSym;
	if ((dlg.DoModal () == IDOK) && (strSym = dlg.m_Symbol.GetBuffer (4)))
	{
		i = StockSymbols.GetIndx (strSym);
		CPosition p (i, StockSymbols.GetPrice (i), dlg.m_NumShares); 
		m_Portfolio.Add (p);
		UpdateAllViews (NULL);
	}
}

//CStockDoc::OnTradeSell - opens a dialog and if the user selects a
// stock and amount it deletes the position from this document. Then 
// it calls update all views to re-display the portfolio.
void CStockDoc::OnTradeSell() 
{
	int i, j;
	CSellDlg dlg;
	char *strSym;

	dlg.m_dpPortfolio = this;
	if ((dlg.DoModal () == IDOK) && (strSym = dlg.m_Symbol.GetBuffer (4)))
	{
		i = StockSymbols.GetIndx (strSym);
		for (j = 0; j < m_Portfolio.GetSize (); j++)
		{
			if (m_Portfolio [j].m_Symbol == i)
			{
				m_Portfolio.RemoveAt (j);
				UpdateAllViews (NULL);
				return;
			}
		}
	}
}

//CStockDoc::OnTradeShort - opens a dialog and if the user selects a
// stock and amount it adds a short position to this document. Then it 
// calls update all views to display the new position.
void CStockDoc::OnTradeShort() 
{
	int i;
	CShortDlg dlg; 
	char *strSym;

	if ((dlg.DoModal () == IDOK) && (strSym = dlg.m_Symbol.GetBuffer (4)))
	{
		i = StockSymbols.GetIndx (strSym);
		CPosition p (i, StockSymbols.GetPrice (i), -dlg.m_NumShares); 
		m_Portfolio.Add (p);
		UpdateAllViews (NULL);
	}
	UpdateAllViews (NULL);
}

