//*
//* symdoc.cpp
//*
//* | Licensed Materials - Use restricted, please refer to the "Samples" terms
//* | and conditions in the IBM International Program License Agreement (IPLA).  
//* (C) Copyright IBM Corporation. 1992, 2006. All Rights Reserved.
//*

// symdoc.cpp : implementation file
//

#include "stdafx.h"
#include "Stock.h"
#include "stockdoc.h"
#include "symdoc.h"

#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CSymDoc

CSymDoc::CSymDoc ()
{
	for (m_Last = 0; m_Last <= MAXTRANSACTIONS; m_Last++)
		m_Prices [m_Last] = m_Volumes [m_Last] = 0;
	m_Last = m_PriceMin = 0;
}

IMPLEMENT_DYNCREATE(CSymDoc, CDocument)

void CSymDoc::Transaction (int PPrice)
{
	m_Last = (m_Last + 1) & MAXTRANSACTIONS;
	m_Prices [m_Last] = PPrice;
	m_Volumes [m_Last] = rand () & 0xf;
	m_PriceMin = max (PPrice - 24, m_PriceMin);
	m_PriceMin = min (PPrice - 8, m_PriceMin);
	GetPriceStr (PPrice, m_StrPrice);
	UpdateAllViews (NULL);
}

//CSymDoc::InitSymbol - sets up a graph view of a stock symbol
// and back generates a trading history.
// Input: PSymbol - string containing stock symbol name.
void CSymDoc::InitSymbol (char* PSymbol)
{
	SetTitle (PSymbol);
	m_Symbol = StockSymbols.GetIndx (PSymbol);
	StockSymbols.SetDoc (m_Symbol, this);
	Transaction (StockSymbols.GetPrice (m_Symbol));
	//generate a trading history
	int p = m_Prices [m_Last];
	for (int indx = (m_Last-1) & MAXTRANSACTIONS; indx != m_Last; 
		indx = (indx - 1) & MAXTRANSACTIONS)
	{
		m_Prices [indx] = p += StockSymbols.GenDelta (m_Symbol);
		m_PriceMin = max (m_Prices [m_Last] - 24, m_PriceMin);
		m_PriceMin = min (m_Prices [m_Last] - 8, m_PriceMin);
  		m_Volumes [indx] = rand () & 0xf;
	}
}

CSymDoc::~CSymDoc()
{
}

BEGIN_MESSAGE_MAP(CSymDoc, CDocument)
	//{{AFX_MSG_MAP(CSymDoc)
		// NOTE - the ClassWizard will add and remove mapping macros here.
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


/////////////////////////////////////////////////////////////////////////////
// CSymDoc diagnostics

#ifdef _DEBUG
void CSymDoc::AssertValid() const
{
	CDocument::AssertValid();
}

void CSymDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CSymDoc commands

void CSymDoc::OnCloseDocument() 
{
	StockSymbols.SetDoc (m_Symbol, NULL);
	CDocument::OnCloseDocument();
}
